import numpy as np


def harvey(frequency, guesses, mode='regular', total=False, ab=False):
    """
    The main model for the stellar background fitting

    Parameters
    ----------
    frequency : numpy.ndarray
        the frequency of the power spectrum
    guesses : list
        the parameters of the Harvey model
    mode : {'regular', 'second', 'fourth'}
        the mode of which Harvey model parametrization to use. Default mode is `regular`.
        The 'regular' mode is when both the second and fourth order terms are added in the denominator
        whereas, 'second' only adds the second order term and 'fourth' only adds the fourth order term.
    total : bool
        If `True`, returns the summed model over multiple components.

    Returns
    -------
    model : np.ndarray
        the stellar background model

    """
    nlaws = int((len(guesses)-1)/2)
    model = np.zeros_like(frequency)

    if mode == 'regular':
        for i in range(nlaws):
            if not ab:
                model += (4.*(guesses[(i*2)+1]**2.)*guesses[i*2])/(1.0+(2.*np.pi*guesses[i*2]*frequency)**2.0+(2.*np.pi*guesses[i*2]*frequency)**4.0)
            else:
                model += guesses[i*2]/(1.0+(guesses[(i*2)+1]*frequency)**2.0+(guesses[(i*2)+1]*frequency)**4.0)
    elif mode == 'second':
        for i in range(nlaws):
            if not ab:
                model += (4.*(guesses[(i*2)+1]**2.)*guesses[i*2])/(1.0+(2.*np.pi*guesses[i*2]*frequency)**2.0)
            else:
                model += guesses[i*2]/(1.0+(guesses[(i*2)+1]*frequency)**2.0)
    elif mode == 'fourth':
        for i in range(nlaws):
            if not ab:
                model += (4.*(guesses[(i*2)+1]**2.)*guesses[i*2])/(1.0+(2.*np.pi*guesses[i*2]*frequency)**4.0)
            else:
                model += guesses[i*2]/(1.0+(guesses[(i*2)+1]*frequency)**4.0)
    else:
        pass

    if total:
        model += guesses[2*nlaws]

    return model


def gaussian(frequency, offset, amplitude, center, width):
    """
    The Gaussian function.

    Parameters
    ----------
    frequency : numpy.ndarray
        the frequency array
    offset : float
        the vertical offset
    amplitude : float
        amplitude of the Gaussian
    center : float
        center of the Gaussian
    width : float
        the width of the Gaussian

    Returns
    -------
    result : np.ndarray
        the Gaussian function

    """

    model = np.zeros_like(frequency)

    model += amplitude*np.exp(-(center-frequency)**2.0/(2.0*width**2))
    model += offset

    return model


def harvey_none(frequency, white_noise, ab=False, fix=False):
    """
    No Harvey model

    Parameters
    ----------
    frequency : numpy.ndarray
        the frequency array
    white_noise : float
        the white noise component

    Returns
    -------
    model : numpy.ndarray
        the no-Harvey (white noise) model

    """

    model = np.zeros_like(frequency)
    if not fix:
        model += white_noise

    return model


def harvey_one(frequency, tau_1, sigma_1, white_noise, ab=False, fix=False):
    """
    One Harvey model

    Parameters
    ----------
    frequency : numpy.ndarray
        the frequency array
    tau_1 : float
        timescale of the first harvey component [Ms]
    sigma_1 : float
        amplitude of the first harvey component
    white_noise : float
        the white noise component

    Returns
    -------
    model : numpy.ndarray
        the one-Harvey model

    """

    model = np.zeros_like(frequency)
    if not ab:
        model += (4.*(sigma_1**2.)*tau_1)/(1.0+(2.*np.pi*tau_1*frequency)**2.0+(2.*np.pi*tau_1*frequency)**4.0)
    else:
        model += tau_1/(1.0+(sigma_1*frequency)**2.0+(sigma_1*frequency)**4.0)
    if not fix:
        model += white_noise

    return model


def harvey_two(frequency, tau_1, sigma_1, tau_2, sigma_2, white_noise, ab=False, fix=False):
    """
    Two Harvey model

    Parameters
    ----------
    frequency : numpy.ndarray
        the frequency array
    tau_1 : float
        timescale of the first harvey component
    sigma_1 : float
        amplitude of the first harvey component
    tau_2 : float
        timescale of the second harvey component
    sigma_2 : float
        amplitude of the second harvey component
    white_noise : float
        the white noise component

    Returns
    -------
    model : numpy.ndarray
        the two-Harvey model

    """

    model = np.zeros_like(frequency)
    if not ab:
        model += (4.*(sigma_1**2.)*tau_1)/(1.0+(2.*np.pi*tau_1*frequency)**2.0+(2.*np.pi*tau_1*frequency)**4.0)
        model += (4.*(sigma_2**2.)*tau_2)/(1.0+(2.*np.pi*tau_2*frequency)**2.0+(2.*np.pi*tau_2*frequency)**4.0)
    else:
        model += tau_1/(1.0+(sigma_1*frequency)**2.0+(sigma_1*frequency)**4.0)
        model += tau_2/(1.0+(sigma_2*frequency)**2.0+(sigma_2*frequency)**4.0)
    if not fix:
        model += white_noise

    return model


def harvey_three(frequency, tau_1, sigma_1, tau_2, sigma_2, tau_3, sigma_3, white_noise, ab=False, fix=False):
    """
    Three Harvey model

    Parameters
    ----------
    frequency : numpy.ndarray
        the frequency array
    tau_1 : float
        timescale of the first harvey component
    sigma_1 : float
        amplitude of the first harvey component
    tau_2 : float
        timescale of the second harvey component
    sigma_2 : float
        amplitude of the second harvey component
    tau_3 : float
        timescale of the third harvey component
    sigma_3 : float
        amplitude of the third harvey component
    white_noise : float
        the white noise component

    Returns
    -------
    model : numpy.ndarray
        the three-Harvey model
    """

    model = np.zeros_like(frequency)
    if not ab:
        model += (4.*(sigma_1**2.)*tau_1)/(1.0+(2.*np.pi*tau_1*frequency)**2.0+(2.*np.pi*tau_1*frequency)**4.0)
        model += (4.*(sigma_2**2.)*tau_2)/(1.0+(2.*np.pi*tau_2*frequency)**2.0+(2.*np.pi*tau_2*frequency)**4.0)
        model += (4.*(sigma_3**2.)*tau_3)/(1.0+(2.*np.pi*tau_3*frequency)**2.0+(2.*np.pi*tau_3*frequency)**4.0)
    else:
        model += tau_1/(1.0+(sigma_1*frequency)**2.0+(sigma_1*frequency)**4.0)
        model += tau_2/(1.0+(sigma_2*frequency)**2.0+(sigma_2*frequency)**4.0)
        model += tau_3/(1.0+(sigma_3*frequency)**2.0+(sigma_3*frequency)**4.0)
    if not fix:
        model += white_noise

    return model


import numpy as np
import pandas as pd
from collections import deque
from astropy.convolution import Box1DKernel, Gaussian1DKernel, convolve, convolve_fft



def set_seed(star, lower=1, upper=10**7, size=1):
    """
    For Kepler targets that require a correction via CLI (--kc), a random seed is generated
    from U~[1,10^7] and stored in stars_info.csv for reproducible results in later runs.

    Parameters
    ----------
    star : target.Target
        the pySYD pipeline object
    lower : int 
        lower limit for random seed value. Default value is `1`.
    upper : int
        upper limit for random seed value. Default value is `10**7`.
    size : int
        number of seed values returned. Default value is `1`.

    Returns
    -------
    star : target.Target
        the pySYD pipeline object
        
    """

    seed = list(np.random.randint(lower,high=upper,size=size))
    df = pd.read_csv(star.params['info'])
    stars = [str(each) for each in df.stars.values.tolist()]
    idx = stars.index(star.name)
    df.loc[idx,'seed'] = int(seed[0])
    star.params[star.name]['seed'] = seed[0]
    df.to_csv(star.params['info'],index=False)
    return star


def remove_artefact(star, lcp=1.0/(29.4244*60*1e-6), lf_lower=[240.0,500.0], lf_upper =[380.0,530.0], 
                    hf_lower = [4530.0,5011.0,5097.0,5575.0,7020.0,7440.0,7864.0],
                    hf_upper = [4534.0,5020.0,5099.0,5585.0,7030.0,7450.0,7867.0],):
    """
    Removes SC artefacts in Kepler power spectra by replacing them with noise (using linear interpolation)
    following a chi-squared distribution. 

    Known artefacts are:
    1) 1./LC harmonics
    2) high frequency artefacts (>5000 muHz)
    3) low frequency artefacts 250-400 muHz (mostly present in Q0 and Q3 data)

    Parameters
    ----------
    star : target.Target
        the pySYD pipeline object
    lcp : float
        long cadence period in Msec
    lf_lower : List[float]
        lower limit of low frequency artefact
    lf_upper : List[float]
        upper limit of low frequency artefact
    hf_lower : List[float]
        lower limit of high frequency artefact
    hf_upper : List[float]
        upper limit of high frequency artefact
    Returns
    -------
    star : target.Target
        the pySYD pipeline object
    """

    if star.params[star.name]['seed'] is None:
        star = set_seed(star)
    # LC period in Msec -> 1/LC ~muHz
    artefact = (1.0+np.arange(14))*lcp
    # Estimate white noise
    white = np.mean(star.power[(star.frequency >= star.nyquist-100.0)&(star.frequency <= star.nyquist-50.0)])

    np.random.seed(int(star.params[star.name]['seed']))
    # Routine 1: remove 1/LC artefacts by subtracting +/- 5 muHz given each artefact
    for i in range(len(artefact)):
        if artefact[i] < np.max(star.frequency):
            mask = np.ma.getmask(np.ma.masked_inside(star.frequency, artefact[i]-5.0*star.resolution, artefact[i]+5.0*star.resolution))
            if np.sum(mask) != 0:
                star.power[mask] = white*np.random.chisquare(2,np.sum(mask))/2.0

    np.random.seed(int(star.params[star.name]['seed']))
    # Routine 2: fix high frequency artefacts
    for lower, upper in zip(hf_lower, hf_upper):
        if lower < np.max(star.frequency):
            mask = np.ma.getmask(np.ma.masked_inside(star.frequency, lower, upper))
            if np.sum(mask) != 0:
                star.power[mask] = white*np.random.chisquare(2,np.sum(mask))/2.0

    np.random.seed(int(star.params[star.name]['seed']))
    # Routine 3: remove wider, low frequency artefacts 
    for lower, upper in zip(lf_lower, lf_upper):
        low = np.ma.getmask(np.ma.masked_outside(star.frequency, lower-20., lower))
        upp = np.ma.getmask(np.ma.masked_outside(star.frequency, upper, upper+20.))
        # Coeffs for linear fit
        m, b = np.polyfit(star.frequency[~(low*upp)], star.power[~(low*upp)], 1)
        mask = np.ma.getmask(np.ma.masked_inside(star.frequency, lower, upper))
        # Fill artefact frequencies with noise
        star.power[mask] = ((star.frequency[mask]*m)+b)*(np.random.chisquare(2, np.sum(mask))/2.0)

    return star


def whiten_mixed(star):
    """
    Generates random white noise in place of ell=1 for subgiants with mixed modes to better
    constrain the characteristic frequency spacing.

    Parameters
    ----------
    star : target.Target
        pySYD pipeline target
    star.frequency : np.ndarray
        the frequency of the power spectrum
    star.power : np.ndarray
        the power spectrum

    """
    if star.params[star.name]['seed'] is None:
        star = set_seed(star)
    # Estimate white noise
    white = np.mean(star.power[(star.frequency >= star.nyquist-100.0)&(star.frequency <= star.nyquist-50.0)])

    # Take the provided dnu and "fold" the power spectrum
    folded_freq = np.copy(star.frequency)%star.params[star.name]['dnu']
    mask = np.ma.getmask(np.ma.masked_inside(folded_freq, star.params[star.name]['ech_mask'][0], star.params[star.name]['ech_mask'][1]))
    np.random.seed(int(star.params[star.name]['seed']))
    # Routine 1: remove 1/LC artefacts by subtracting +/- 5 muHz given each artefact
    if np.sum(mask) != 0:
        star.power[mask] = white*np.random.chisquare(2,np.sum(mask))/2.0

    return star


def log_likelihood(observations, model):
    """
    Until we figure out a better method, we are computing the likelhood using
    the mean squared error.

    Parameters
    ----------
    observations : ndarray
        the observed power spectrum
    model : ndarray
        model generated at the observed frequencies

    Returns
    -------
    LL : float
        the natural logarithm of the likelihood (or the MSE)

    """

    return -0.5*(np.sum((observations-model)**2.))


def compute_aic(observations, model, n_parameters):
    """
    Computes the Akaike Information Criterion (AIC) given the modeled
    power spectrum.

    Parameters
    ----------
    observations : ndarray
        the observed power spectrum
    model : ndarray
        model generated at the observed frequencies
    n_parameters : int
        number of free parameters in the given model

    Returns
    -------
    aic : float
        AIC value

    """
    N = len(observations)
    LL = log_likelihood(observations, model)
    aic = (-2.*LL)/N + (2.*n_parameters)/N

    return aic


def compute_bic(observations, model, n_parameters):
    """
    Computes the Bayesian Information Criterion (BIC) given the modeled 
    power spectrum.

    Parameters
    ----------
    observations : ndarray
        the observed power spectrum
    model : ndarray
        model generated at the observed frequencies
    n_parameters : int
        number of free parameters in the given model

    Returns
    -------
    aic : float
        AIC value

    """
    N = len(observations)
    LL = log_likelihood(observations, model)
    bic = -2.*LL + np.log(N)*n_parameters

    return bic


def gaussian_bounds(x, y, guesses, best_x=None, sigma=None):
    """
    Get the bounds for the parameters of a Gaussian fit to the data.

    Parameters
    ----------
    x : np.ndarray
        the x values of the data
    y : np.ndarray
        the y values of the data
    guesses : List[float]
        initial guesses for a gaussian fit
    sigma : Optional[float]
        sigma from a standard Normal distribution 

    Returns
    -------
    bb : List[Tuple]
        list of parameter bounds of a Gaussian fit to the data
    """

    offset, amp, center, width = guesses
    if sigma is None:
        sigma = (max(x)-min(x))/8.0/np.sqrt(8.0*np.log(2.0))
    bb = []
    b = np.zeros((2, 4)).tolist()
    
    # offset bound:
    b[1][0] = np.inf            #upper bound
    # amplitude bounds:
    if amp > 0:
        b[1][1] = 2.0*np.max(y) #upper bound
    else:
        b[0][1]=-np.inf         #lower bound
        b[1][1]=0               #upper bound
    # center bounds:
    b[0][2] = np.min(x)
    b[1][2] = np.max(x)
    # width bounds
    b[0][3] = sigma
    b[1][3] = (np.max(x)-np.min(x))*2.
    bb.append(tuple(b))
    
    return bb


def max_elements(x, y, npeaks):
    """
    Get the first n peaks of the given data.

    Parameters
    ----------
    x : np.ndarray
        the x values of the data
    y : np.ndarray
        the y values of the data
    npeaks : int
        the first n peaks

    Returns
    -------
    peaks_x : np.ndarray
        the x co-ordinates of the first `npeaks`
    peaks_y : np.ndarray
        the y co-ordinates of the first `npeaks`
    """

    s = np.argsort(y)
    peaks_y = y[s][-int(npeaks):][::-1]
    peaks_x = x[s][-int(npeaks):][::-1]

    return peaks_x, peaks_y


def return_max(x_array, y_array, exp_dnu=None, index=False):
    """
    Return the either the value of peak or the index of the peak corresponding to the most likely dnu given a prior estimate,
    otherwise just the maximum value.

    Parameters
    ----------
    x_array : np.ndarray
        the independent axis (i.e. time, frequency)
    y_array : np.ndarray
        the dependent axis
    method : str
        which method to use for determing the max elements in an array
    index : bool
        if true will return the index of the peak instead otherwise it will return the value. Default value is `False`.
    dnu : bool
        if true will choose the peak closest to the expected dnu `exp_dnu`. Default value is `False`.
    exp_dnu : Required[float]
        the expected dnu. Default value is `None`.

    Returns
    -------
    result : Union[int, float]
        if `index` is `True`, result will be the index of the peak otherwise if `index` is `False` it will instead return the
        value of the peak.

    """
    if exp_dnu is None:
        lst = list(y_array)
        idx = lst.index(max(lst))
        weights = None
    else:
        sig = 0.35*exp_dnu/2.35482 
        weights = 1./(sig*np.sqrt(2.*np.pi))*np.exp(-(x_array-exp_dnu)**2./(2.*sig**2))
        lst = list(weights*y_array)
        idx = lst.index(max(lst))
    if index:
        return idx
    else:
        return x_array[idx], y_array[idx]


def bin_data(x, y, width, log=False, mode='mean'):
    """
    Bins a series of data.

    Parameters
    ----------
    x : np.ndarray
        the x values of the data
    y : np.ndarray
        the y values of the data
    width : float
        bin width in muHz
    log : bool
        creates bins by using the log of the min/max values (i.e. not equally spaced in log if `True`)

    Returns
    -------
    bin_x : np.ndarray
        binned frequencies
    bin_y : np.ndarray
        binned power
    bin_yerr : numpy.ndarray
        standard deviation of the binned y data

    """
    if log:
        mi = np.log10(min(x))
        ma = np.log10(max(x))
        no = np.int(np.ceil((ma-mi)/width))
        bins = np.logspace(mi, mi+(no+1)*width, no)
    else:
        bins = np.arange(min(x), max(x)+width, width)

    digitized = np.digitize(x, bins)
    if mode == 'mean':
        bin_x = np.array([x[digitized == i].mean() for i in range(1, len(bins)) if len(x[digitized == i]) > 0])
        bin_y = np.array([y[digitized == i].mean() for i in range(1, len(bins)) if len(x[digitized == i]) > 0])
    elif mode == 'median':
        bin_x = np.array([np.median(x[digitized == i]) for i in range(1, len(bins)) if len(x[digitized == i]) > 0])
        bin_y = np.array([np.median(y[digitized == i]) for i in range(1, len(bins)) if len(x[digitized == i]) > 0])
    else:
        pass
    bin_yerr = np.array([y[digitized == i].std()/np.sqrt(len(y[digitized == i])) for i in range(1, len(bins)) if len(x[digitized == i]) > 0])

    return bin_x, bin_y, bin_yerr


def delta_nu(numax):
    """
    Estimates dnu using numax scaling relation.

    Parameters
    ----------
    numax : float
        the estimated numax

    Returns
    -------
    dnu : float
        the estimated dnu

    """

    return 0.22*(numax**0.797)




import pickle
import numpy as np
from tqdm import tqdm
import matplotlib.pyplot as plt
from scipy.signal import find_peaks
from scipy.optimize import curve_fit
from scipy.interpolate import InterpolatedUnivariateSpline
from astropy.convolution import Box1DKernel, Gaussian1DKernel, convolve, convolve_fft

from pysyd import functions
from pysyd import models
from pysyd import utils
from pysyd import plots



class Target:

    def __init__(self, star, args):
        """
        A pySYD pipeline target. Initialization stores all the relevant information and
        checks/loads in data for the given target. pySYD no longer requires BOTH the time
        series data and the power spectrum, but requires additional information via CLI if
        the former is not provided i.e. cadence or nyquist frequency, the oversampling
        factor (if relevant), etc.
    
        Attributes
        ----------
        star : int
            the star ID
        params : Dict[str,object]
            the pipeline parameters
        findex : Dict[str,object]
            the parameters of the find excess routine
        fitbg : Dict[str,object]
            the parameters relevant for the background-fitting procedure
        globe : Dict[str,object]
            parameters relevant for estimating global asteroseismic parameters numax and dnu
        verbose : bool
            if true, turns on the verbose output
    
        Parameters
        ----------
        args : argparse.Namespace
            the parsed and updated command line arguments

        Methods
        -------
        TODO: Add methods

        """
        self.name = star
        self.params = args.params
        self.findex = args.findex
        self.fitbg = args.fitbg
        self.globe = args.globe
        self.verbose = args.verbose
        self = utils.load_data(self, args)


    def run_syd(self):
        """
        Run the pySYD pipeline routines sequentially:
        1) the find excess module to identify the any solar-like oscillations
        2) estimates the stellar background contributions before estimating the
           global asteroseismic parameters

        Returns
        -------
        None

        """
        # Run the find excess routine
        if self.params[self.name]['excess']:
            self = utils.get_findex(self)
            self.find_excess()
        # Run the global fitting routine
        if utils.check_fitbg(self):
            self = utils.get_fitbg(self)
            self.fit_global()
        if self.params['show']:
            note=''
            if self.verbose:
                note+=' - displaying figures'
            print(note)
            plt.show(block=False)
            input(' - press RETURN to exit')
            if not self.verbose:
                print('')


    def find_excess(self):
        """
        Automatically finds power excess due to solar-like oscillations using a
        frequency-resolved, collapsed autocorrelation function (ACF).

        Returns
        -------
        None

        """
        # Make sure the binning is specified, otherwise it cannot run
        if self.findex['binning'] is not None:
            # Smooth the power in log-space
            self.bin_freq, self.bin_pow, self.bin_pow_err = functions.bin_data(self.freq, self.pow, width=self.findex['binning'], log=True, mode=self.findex['mode'])
            # Smooth the power in linear-space
            self.smooth_freq, self.smooth_pow, self.smooth_pow_err = functions.bin_data(self.bin_freq, self.bin_pow, width=self.findex['smooth_width'])
            if self.verbose:
                print('------------------------------------------------------')
                print('Running find_excess module:')
                print('PS binned to %d datapoints' % len(self.smooth_freq))

            # Interpolate and divide to get a crude background-corrected power spectrum
            s = InterpolatedUnivariateSpline(self.smooth_freq, self.smooth_pow, k=1)
            self.interp_pow = s(self.freq)
            self.bgcorr_pow = self.pow/self.interp_pow

            # Calculate collapsed ACF using different box (or bin) sizes
            self.findex['results'][self.name] = {}
            self.compare = []
            for b in range(self.findex['n_trials']):
                self.collapsed_acf(b)

            # Select trial that resulted with the highest SNR detection
            self.findex['results'][self.name]['best'] = self.compare.index(max(self.compare))+1
            if self.verbose:
                print('selecting model %d'%self.findex['results'][self.name]['best'])
            utils.save_findex(self)
            plots.plot_excess(self)
            self.pickles.append('excess.pickle')


    def collapsed_acf(self, b, start=0, max_iterations=5000, max_snr=100.):
        """
        Computes a collapsed autocorrelation function (ACF).

        Parameters
        ----------
        b : int
            the trial number
        start : int
            what index of the frequency array to start with, which is `0` by default.
        max_iterations : int
            maximum number of times to run the scipy.optimization before calling it quits
        j : int
            index at which to start storing the cumulative sum and mean of ACF. Default value is `0`.
        start : int
            index at which to start masking the frequency and power spectrum. Default value is `0`.
        max_iterations : int
            maximum number of interations to try in curve fitting routine. Default value is `5000`.
        max_snr : float
            maximum SNR corresponding to power excess. Default value is `100.0`.

        Returns
        -------
        None

        """
        constants = utils.Constants()
        # Computes a collapsed ACF using different "box" (or bin) sizes
        self.findex['results'][self.name][b+1] = {}
        subset = np.ceil(self.boxes[b]/self.resolution)
        steps = np.ceil((self.boxes[b]*self.findex['step'])/self.resolution)

        cumsum = []
        md = []
        # Iterates through entire power spectrum using box width
        while True:
            if (start+subset) > len(self.freq):
                break
            p = self.bgcorr_pow[int(start):int(start+subset)]
            auto = np.real(np.fft.fft(np.fft.ifft(p)*np.conj(np.fft.ifft(p))))
            cumsum.append(np.sum(np.absolute(auto-np.mean(auto))))
            md.append(np.mean(self.freq[int(start):int(start+subset)]))
            start += steps
        # subtract/normalize the summed ACF and take the max
        md = np.array(md)
        cumsum = np.array(cumsum)-min(cumsum)
        csum = list(cumsum/max(cumsum))
        # Pick the maximum value as an initial guess for numax
        idx = csum.index(max(csum))
        csum = np.array(csum)
        self.findex['results'][self.name][b+1].update({'x':md,'y':csum,'maxx':md[idx],'maxy':csum[idx]})

        # Fit Gaussian to get estimate value for numax
        try:
            best_vars, _ = curve_fit(models.gaussian, md, csum, p0=[np.median(csum), 1.0-np.median(csum), md[idx], constants.width_sun*(md[idx]/constants.numax_sun)], maxfev=max_iterations, bounds=((-np.inf,-np.inf,1,-np.inf),(np.inf,np.inf,np.inf,np.inf)),)
        except Exception as _:
            self.findex['results'][self.name][b+1].update({'good_fit':False})
            snr = 0.
        else:
            self.findex['results'][self.name][b+1].update({'good_fit':True})
            fitx = np.linspace(min(md), max(md), 10000)
            fity = models.gaussian(fitx, *best_vars)
            self.findex['results'][self.name][b+1].update({'fitx':fitx,'fity':fity})
            snr = max(fity)/np.absolute(best_vars[0])
            if snr > max_snr:
                snr = max_snr
            self.findex['results'][self.name][b+1].update({'numax':best_vars[2],'dnu':functions.delta_nu(best_vars[2]),'snr':snr})
            if self.verbose:
                  print('power excess trial %d: numax = %.2f +/- %.2f'%(b+1, best_vars[2], np.absolute(best_vars[3])/2.0))
                  print('S/N: %.2f' % snr)
        self.compare.append(snr)


    def fit_global(self):
        """
        The second main pySYD pipeline routine. First it 
        Perform a fit to the granulation background and measures the frequency of maximum power (numax),
        the large frequency separation (dnu) and oscillation amplitude.

        Returns
        -------
        None

        """
        while self.i < self.fitbg['mc_iter']:
            # Requires convergence of background fit before going to the next step 
            if self.fit_background():
                self.get_numax()
                self.get_dnu()
                # First step?
                if self.i == 0:
                    # Plot results
                    plots.plot_background(self)
                    self.pickles.append('background.pickle')
                    if self.fitbg['mc_iter'] > 1:
                        # Switch to critically-sampled PS if sampling
                        mask = np.ma.getmask(np.ma.masked_inside(self.freq_cs, self.params[self.name]['bg_mask'][0], self.params[self.name]['bg_mask'][1]))
                        self.frequency, self.power = np.copy(self.freq_cs[mask]), np.copy(self.pow_cs[mask])
                        self.resolution = self.frequency[1]-self.frequency[0]
                        if self.verbose:
                            print('------------------------------------------------------\nRunning sampling routine:')
                            self.pbar = tqdm(total=self.fitbg['mc_iter'])
                            self.pbar.update(1)
                else:
                    if self.verbose:
                        self.pbar.update(1)
                self.i += 1
                if self.i == self.fitbg['mc_iter'] and self.fitbg['mc_iter'] > 1 and self.verbose:
                    self.pbar.close()
        # Save results of second module
        utils.save_fitbg(self)
        if self.fitbg['mc_iter'] > 1:
            # Plot results if sampling
            plots.plot_samples(self)
            self.pickles.append('samples.pickle')
        if self.verbose:
            # Print results
            utils.verbose_output(self)


    def fit_background(self):
        """
        Fits the stellar background contribution due to granulation. 

        Returns
        -------
        None

        """
        # Bin power spectrum to model stellar background/correlated red noise components
        if self.i != 0:
            self.random_pow = (np.random.chisquare(2, len(self.frequency))*self.power)/2.
        bin_freq, bin_pow, bin_err = functions.bin_data(self.frequency, self.random_pow, width=self.fitbg['ind_width'], mode=self.findex['mode'])
        # Mask out region with power excess
        mask = np.ma.getmask(np.ma.masked_outside(bin_freq, self.params[self.name]['ps_mask'][0], self.params[self.name]['ps_mask'][1]))
        self.bin_freq = bin_freq[mask]
        self.bin_pow = bin_pow[mask]
        self.bin_err = bin_err[mask]

        if self.i == 0 and self.verbose:
            print('------------------------------------------------------')
            print('Running fit_background module:')
            print('PS binned to %d data points' % len(bin_freq))
        # Estimate white noise level
        self.get_white_noise()
        # Get initial guesses for the optimization of the background model
        self.estimate_initial_red()
        # If optimization does not converge, the rest of the code will not run
        if self.i == 0:
            if self.get_best_model():
                print('WARNING: Bad initial fit for star %d. Check this and try again.'%self.name)
                return False
        else:
            if self.get_red_noise():
                return False
        return True


    def get_white_noise(self):
        """
        Estimate the white noise level (in muHz) by taking the mean of
        the last 10% of the power spectrum.

        Returns
        -------
        None

        """
        mask = (self.frequency > (max(self.frequency)-0.1*max(self.frequency)))&(self.frequency < max(self.frequency))
        self.noise = np.mean(self.random_pow[mask])


    def estimate_initial_red(self, a=[]):
        """
        Estimates amplitude of red noise components by using a smoothed version of the power
        spectrum with the power excess region masked out. This will take the mean of a specified 
        number of points (via -nrms, default=20) for each Harvey-like component.

        Parameters
        ----------
        a : List[float]
            initial guesses for the amplitudes of all Harvey components

        Returns
        -------
        None

        """
        # Exclude region with power excess and smooth to estimate red noise components
        boxkernel = Box1DKernel(int(np.ceil(self.fitbg['box_filter']/self.resolution)))
        mask = (self.frequency >= self.params[self.name]['ps_mask'][0])&(self.frequency <= self.params[self.name]['ps_mask'][1])
        self.smooth_pow = convolve(self.random_pow, boxkernel)
        # Temporary array for inputs into model optimization
        pars = np.zeros((self.nlaws*2+1))
        # Estimate amplitude for each harvey component
        for n, nu in enumerate(self.mnu):
            diff = list(np.absolute(self.frequency-nu))
            idx = diff.index(min(diff))
            if idx < self.fitbg['n_rms']:
                pars[2*n+1] = np.sqrt((np.mean(self.smooth_pow[~mask][:self.fitbg['n_rms']]))/(4.*self.b[n]))
            elif (len(self.smooth_pow[~mask])-idx) < self.fitbg['n_rms']:
                pars[2*n+1] = np.sqrt((np.mean(self.smooth_pow[~mask][-self.fitbg['n_rms']:]))/(4.*self.b[n]))
            else:
                pars[2*n+1] = np.sqrt((np.mean(self.smooth_pow[~mask][idx-int(self.fitbg['n_rms']/2):idx+int(self.fitbg['n_rms']/2)]))/(4.*self.b[n]))
            pars[2*n] = self.b[n]
            a.append(pars[2*n+1])
        pars[-1] = self.noise
        self.pars = pars
        self.a_orig = np.array(a)


    def get_best_model(self):
        """
        Determines the best-fit model for the stellar granulation background in the power spectrum
        by iterating through several models, where the initial guess for the number of Harvey-like 
        component(s) to model is estimated from a solar scaling relation.

        Parameters
        ----------
        names : List[str]
            the number of Harvey components to use in the background model
        bounds : list
            the bounds on the Harvey parameters for a given model
        bic : list
            the BIC statistic
        aic : list
            the AIC statistic
        paras : list
            the Harvey model parameters

        Returns
        -------
        again : bool
            will return `True` if fitting failed and the iteration must be repeated, otherwise `False`.

        """
        # Get best-fit model
        if self.fitbg['n_laws'] is None:
            self.bounds = []
            self.bic = []
            self.aic = []
            self.paras = []
            if self.verbose:
                print('Comparing %d different models:'%(self.nlaws+1))
            for nlaws in range(self.nlaws+1):
                note=''
                bb = np.zeros((2,2*nlaws+1)).tolist()
                if nlaws != 0:
                    for z in range(nlaws):
                        bb[0][int(2*z)] = 0.
                        if self.fitbg['ab']:
                            bb[1][int(2*z)] = np.inf
                        else:
                            bb[1][int(2*z)] = self.baseline/10**6.
                        bb[0][int(2*z+1)] = 0.
                        bb[1][int(2*z+1)] = np.inf
                bb[0][-1] = 0.
                bb[1][-1] = np.inf
                self.bounds.append(tuple(bb))
                if self.verbose:
                    if nlaws == 0:
                        note += 'Model %d: Flat white-noise term only'%nlaws
                    else:
                        note += 'Model %d: White noise term + %d Harvey-like component(s)'%(nlaws,nlaws)
                try:
                    pp, _ = curve_fit(self.fitbg['functions'][nlaws], self.bin_freq, self.bin_pow, p0=self.pars[-int(2*nlaws+1):], sigma=self.bin_err, bounds=tuple(bb))
                except RuntimeError as _:
                    self.paras.append([])
                    self.bic.append(np.inf)
                    self.aic.append(np.inf)
                else:
                    self.paras.append(pp)
                    model = models.harvey(self.bin_freq, pp, total=True)
                    b = functions.compute_bic(self.bin_pow, model, n_parameters=len(pp))
                    self.bic.append(b)
                    a = functions.compute_aic(self.bin_pow, model, n_parameters=len(pp))
                    self.aic.append(a)
#                note += '\n BIC = %.2f | AIC = %.2f'%(b, a)
                    if self.verbose:
                        print(note)
            # If the fitting converged (fix to bic? depending on performance)
            if self.fitbg['metric'] == 'bic':
                if np.isfinite(min(self.bic)):
                    self.save_best_model()
                    return False
                else:
                    return True
            else:
                if np.isfinite(min(self.aic)):
                    self.save_best_model()
                    return False
                else:
                    return True
        else:
            note=''
            bb = np.zeros((2,2*self.fitbg['n_laws']+1)).tolist()
            if self.fitbg['n_laws'] != 0:
                for z in range(self.fitbg['n_laws']):
                    bb[0][int(2*z)] = 0.
                    bb[1][int(2*z)] = self.baseline/10**6.
                    bb[0][int(2*z+1)] = 0.
                    bb[1][int(2*z+1)] = np.inf
            bb[0][-1] = 0.
            bb[1][-1] = np.inf
            self.bounds = tuple(bb)
            if self.verbose:
                if nlaws == 0:
                    note += 'Using flat white-noise term only'%self.fitbg['n_laws']
                else:
                    note += 'Using %d Harvey-like component(s) + flat white-noise term'%self.fitbg['n_laws']
            try:
                pp, _ = curve_fit(self.fitbg['functions'][self.fitbg['n_laws']], self.bin_freq, self.bin_pow, p0=self.pars[-int(2*self.fitbg['n_laws']+1):], sigma=self.bin_err, bounds=tuple(bb))
            except RuntimeError as _:
                self.pars = []
            else:
                self.pars = pp
            if self.verbose:
                print(note)
            # If the fitting converged (fix to bic? depending on performance)
            if self.pars != []:
                self.save_best_model()
                return False
            else:
                return True


    def save_best_model(self, use='bic'):
        """
        Saves information re: the selected best-fit model (for the stellar background).
        
        Parameters
        ----------
        use : str
            which metric to use for model selection, choices ~['bic','aic']. Default is `'bic'`.

        """
        if self.fitbg['n_laws'] is None:
            if self.fitbg['metric'] == 'bic':
                model = self.bic.index(min(self.bic))
            else:
                model = self.aic.index(min(self.aic))
            # Store model results for plotting
            if self.nlaws != model:
                self.nlaws = model
                self.b = self.b[:(self.nlaws)]
                self.mnu = self.mnu[:(self.nlaws)]
            if self.verbose:
                print('Based on %s statistic: model %d'%(use.upper(),model))
            self.bounds = self.bounds[model]
            self.pars = self.paras[model]
        else:
            if self.nlaws != self.fitbg['n_laws']:
                self.nlaws = self.fitbg['n_laws']
                self.b = self.b[:(self.nlaws)]
        self.bg_corr = self.random_pow/models.harvey(self.frequency, self.pars, total=True)
        # Save background-corrected power spectrum
        utils.save_file(self)
        # Create appropriate keys for star based on nlaws
        for n in range(self.nlaws):
            self.fitbg['results'][self.name]['tau_%d'%(n+1)] = []
            self.fitbg['results'][self.name]['sigma_%d'%(n+1)] = []
        # Save the final values
        for n in range(self.nlaws):
            self.fitbg['results'][self.name]['tau_%d'%(n+1)].append(self.pars[2*n]*10**6.)
            self.fitbg['results'][self.name]['sigma_%d'%(n+1)].append(self.pars[2*n+1])
        self.fitbg['results'][self.name]['white'].append(self.pars[2*self.nlaws])


    def get_numax(self):
        """
        Simple function to call both numax methods.

        Returns
        -------
        None

        """
        self.get_numax_smooth()
        self.get_numax_gaussian()


    def get_numax_smooth(self, divide=True):
        """
        Estimate numax by smoothing the power spectrum and taking the peak. Also
        computes the background-corrected power spectrum and saves to a text file.

        Returns
        -------
        None

        """
        constants = utils.Constants()
        # Smoothing width for determining numax
        if self.globe['sm_par'] is not None:
            self.sm_par = self.globe['sm_par']
        else:
            self.sm_par = 4.*(self.params[self.name]['numax']/constants.numax_sun)**0.2
            if self.sm_par < 1.:
                self.sm_par = 1.
        sig = (self.sm_par*(self.params[self.name]['dnu']/self.resolution))/np.sqrt(8.0*np.log(2.0))
        self.pssm = convolve_fft(np.copy(self.random_pow), Gaussian1DKernel(int(sig)))
        self.pssm_bgcorr = self.pssm-models.harvey(self.frequency, self.pars, total=True)
        mask = np.ma.getmask(np.ma.masked_inside(self.frequency, self.params[self.name]['ps_mask'][0], self.params[self.name]['ps_mask'][1]))
        self.region_freq = self.frequency[mask]
        self.region_pow = self.pssm_bgcorr[mask]
        idx = functions.return_max(self.region_freq, self.region_pow, index=True)
        self.fitbg['results'][self.name]['numax_smooth'].append(self.region_freq[idx])
        self.fitbg['results'][self.name]['A_smooth'].append(self.region_pow[idx])
        # Initial guesses for the parameters of the Gaussian fit to the power envelope
        self.guesses = [0.0, max(self.region_pow), self.region_freq[idx], (max(self.region_freq)-min(self.region_freq))/np.sqrt(8.0*np.log(2.0))]


    def get_numax_gaussian(self, maxfev=5000):
        """
        Estimate numax by fitting a Gaussian to the power envelope of the smoothed power spectrum.
    
        Parameters
        ----------
        maxfev : int, optional
            maximum number of attempts for the scipy.curve_fit optimization step

        Returns
        -------
        None

        """
        constants = utils.Constants()
        bb = functions.gaussian_bounds(self.region_freq, self.region_pow, self.guesses)
        p_gauss1, _ = curve_fit(models.gaussian, self.region_freq, self.region_pow, p0=self.guesses, bounds=bb[0], maxfev=maxfev)
        # create array with finer resolution for purposes of quantifying uncertainty
        self.new_freq = np.linspace(min(self.region_freq), max(self.region_freq), 10000)
        self.numax_fit = list(models.gaussian(self.new_freq, *p_gauss1))
        d = self.numax_fit.index(max(self.numax_fit))
        self.fitbg['results'][self.name]['numax_gauss'].append(self.new_freq[d])
        self.fitbg['results'][self.name]['A_gauss'].append(p_gauss1[1])
        self.fitbg['results'][self.name]['FWHM'].append(p_gauss1[3])


    def compute_acf(self, fft=True):
        """
        Compute the ACF of the smooth background corrected power spectrum.

        Parameters
        ----------
        fft : bool
            if true will use FFT to compute the ACF. Default value is `True`.

        Returns
        -------
        None

        """
        # Optional smoothing of PS to remove fine structure before computing ACF
        if self.globe['smooth_ps'] is not None:
            boxkernel = Box1DKernel(int(np.ceil(self.globe['smooth_ps']/self.resolution)))
            self.bg_corr_smooth = convolve(self.bg_corr, boxkernel)
        else:
            self.bg_corr_smooth = np.copy(self.bg_corr)
        # Use only power near the expected numax to reduce additional noise in ACF
        power = self.bg_corr_smooth[(self.frequency >= self.params[self.name]['ps_mask'][0])&(self.frequency <= self.params[self.name]['ps_mask'][1])]
        lag = np.arange(0.0, len(power))*self.resolution
        if fft:
            auto = np.real(np.fft.fft(np.fft.ifft(power)*np.conj(np.fft.ifft(power))))
        else:
            auto = np.correlate(power-np.mean(power), power-np.mean(power), "full")
            auto = auto[int(auto.size/2):]
        mask = np.ma.getmask(np.ma.masked_inside(lag, self.params[self.name]['dnu']/4., 2.*self.params[self.name]['dnu']+self.params[self.name]['dnu']/4.))
        lag = lag[mask]
        auto = auto[mask]
        auto -= min(auto)
        auto /= max(auto)
        self.lag = np.copy(lag)
        self.auto = np.copy(auto)


    def get_acf_cutout(self, threshold=1.0):
        """
        Estimate the large frequency spacing or dnu.
        NOTE: this is only used during the first iteration!

        Parameters
        ----------
        threshold : float
            the threshold is multiplied by the full-width half-maximum value, centered on the peak 
            in the ACF to determine the width of the cutout region.
        
        Returns
        -------
        None

        """
        # Get peaks from ACF
        peak_idx,_ = find_peaks(self.auto)
        peaks_l,peaks_a = self.lag[peak_idx],self.auto[peak_idx]
        
        # Pick n highest peaks
        peaks_l = peaks_l[peaks_a.argsort()[::-1]][:self.globe['n_peaks']]
        peaks_a = peaks_a[peaks_a.argsort()[::-1]][:self.globe['n_peaks']]
        # Pick best peak in ACF by using Gaussian weight according to expected dnu
        idx = functions.return_max(peaks_l, peaks_a, index=True, exp_dnu=self.params[self.name]['dnu'])
        self.best_lag = peaks_l[idx]
        self.best_auto = peaks_a[idx]
        # Change fitted value with nan to highlight differently in plot
        peaks_l[idx] = np.nan
        peaks_a[idx] = np.nan
        self.peaks_l = peaks_l
        self.peaks_a = peaks_a
        # Calculate FWHM
        if list(self.lag[(self.lag<self.best_lag)&(self.auto<=self.best_auto/2.)]) != []:
            left_lag = self.lag[(self.lag<self.best_lag)&(self.auto<=self.best_auto/2.)][-1]
            left_auto = self.auto[(self.lag<self.best_lag)&(self.auto<=self.best_auto/2.)][-1]
        else:
            left_lag = self.lag[0]
            left_auto = self.auto[0]
        if list(self.lag[(self.lag>self.best_lag)&(self.auto<=self.best_auto/2.)]) != []:
            right_lag = self.lag[(self.lag>self.best_lag)&(self.auto<=self.best_auto/2.)][0]
            right_auto = self.auto[(self.lag>self.best_lag)&(self.auto<=self.best_auto/2.)][0]
        else:
            right_lag = self.lag[-1]
            right_auto = self.auto[-1]
        # Lag limits to use for ACF mask or "cutout"
        self.params[self.name]['acf_mask']=[self.best_lag-(self.globe['threshold']*((right_lag-left_lag)/2.)),self.best_lag+(self.globe['threshold']*((right_lag-left_lag)/2.))]
        self.zoom_lag = self.lag[(self.lag>=self.params[self.name]['acf_mask'][0])&(self.lag<=self.params[self.name]['acf_mask'][1])]
        self.zoom_auto = self.auto[(self.lag>=self.params[self.name]['acf_mask'][0])&(self.lag<=self.params[self.name]['acf_mask'][1])]
        # Boundary conditions and initial guesses stay the same for all iterations
        self.acf_guesses = [np.mean(self.zoom_auto), self.best_auto, self.best_lag, self.best_lag*0.01*2.]
        self.acf_bb = functions.gaussian_bounds(self.zoom_lag, self.zoom_auto, self.acf_guesses, best_x=self.best_lag, sigma=10**-2)
        # Fit a Gaussian function to the selected peak in the ACF to get dnu
        p_gauss3, _ = curve_fit(models.gaussian, self.zoom_lag, self.zoom_auto, p0=self.acf_guesses, bounds=self.acf_bb[0])
       	# If dnu is provided, use that instead
        if self.params[self.name]['force']:
            p_gauss3[2] = self.params[self.name]['guess']
        self.fitbg['results'][self.name]['dnu'].append(p_gauss3[2])
        # Save for plotting
        self.new_lag = np.linspace(min(self.zoom_lag),max(self.zoom_lag),2000)
        self.dnu_fit = models.gaussian(self.new_lag,*p_gauss3)
        self.obs_acf = max(self.dnu_fit)


    def get_dnu(self):
        """
        Estimate a value for dnu.

        Returns
        -------
        None

        """
        self.compute_acf()
        if self.i == 0:
            self.get_acf_cutout()
            self.get_ridges()
        else:
            # define the peak in the ACF
            zoom_lag = self.lag[(self.lag>=self.params[self.name]['acf_mask'][0])&(self.lag<=self.params[self.name]['acf_mask'][1])]
            zoom_auto = self.auto[(self.lag>=self.params[self.name]['acf_mask'][0])&(self.lag<=self.params[self.name]['acf_mask'][1])]
            # fit a Gaussian function to the selected peak in the ACF
            p_gauss3, _ = curve_fit(models.gaussian, zoom_lag, zoom_auto, p0=self.acf_guesses, bounds=self.acf_bb[0])
            # the center of that Gaussian is our estimate for Dnu
            self.fitbg['results'][self.name]['dnu'].append(p_gauss3[2]) 


    def get_ridges(self, start=0.0):
        """
        Create echelle diagram.

        Parameters
        ----------
        start : float
            lower limit of distance modulus. Default value is `0.0`.

        Returns
        -------
        None

        """
        ech, gridx, gridy, extent = self.echelle()
        N, M = ech.shape[0], ech.shape[1]
        ech_copy = np.array(list(ech.reshape(-1)))

        n = int(np.ceil(self.fitbg['results'][self.name]['dnu'][0]/self.resolution))
        xax = np.zeros(n)
        yax = np.zeros(n)
        modx = self.frequency%self.fitbg['results'][self.name]['dnu'][0]
        for k in range(n):
            use = np.where((modx >= start)&(modx < start+self.resolution))[0]
            if len(use) == 0:
                continue
            xax[k] = np.median(modx[use])
            yax[k] = np.sum(self.bg_corr[use])
            start += self.resolution
        xax = np.array(list(xax)+list(xax+self.fitbg['results'][self.name]['dnu'][0]))
        yax = np.array(list(yax)+list(yax))-min(yax)
        mask = np.ma.getmask(np.ma.masked_where(yax == 0.0, yax))
        # Clip the lower bound (`clip_value`)
        if self.globe['clip_ech']:
            if self.globe['clip_value'] is not None:
                cut = self.globe['clip_value']
            else:
                cut = np.nanmedian(ech_copy)+(3.0*np.nanmedian(ech_copy))
            ech_copy[ech_copy > cut] = cut
        self.ech_copy = ech_copy
        self.ech = ech_copy.reshape((N, M))
        self.extent = extent
        self.xax = xax[~mask]
        self.yax = yax[~mask]


    def echelle(self, n_across=50, startx=0.0):
        """
        Creates an echelle diagram.

        Parameters
        ----------
        n_across : int
            number of grid points in x-axis of echelle diagram. Default value is `50`.
        startx : float
            lower limit of distance modulus. Default value is `0.0`.

        Returns
        -------
        smoothed : numpy.meshgrid
            resulting echelle diagram based on the observed $\delta \nu$ 
        gridx : numpy.ndarray
            grid of x-axis measurements (distance modulus) for echelle diagram
        gridy : numpy.ndarray
            grid of y-axis measurements (frequency) for echelle diagram
        extent : List[float]
            The bounding box in data coordinates that the image will fill. 
            The image is stretched individually along x and y to fill the box.

        """
        if self.globe['smooth_ech'] is not None:
            boxkernel = Box1DKernel(int(np.ceil(self.globe['smooth_ech']/self.resolution)))
            smooth_y = convolve(self.bg_corr, boxkernel)
        nox = self.globe['n_across']
        # fix ndown (no obvious way to do this yet)
        noy = int(np.ceil((max(self.frequency)-min(self.frequency))/self.fitbg['results'][self.name]['dnu'][0]))
        if nox > 2 and noy > 5:
            xax = np.arange(0.0, self.fitbg['results'][self.name]['dnu'][0]+(self.fitbg['results'][self.name]['dnu'][0]/nox)/2.0, self.fitbg['results'][self.name]['dnu'][0]/nox)
            yax = np.arange(min(self.frequency), max(self.frequency), self.fitbg['results'][self.name]['dnu'][0])
            arr = np.zeros((len(xax), len(yax)))
            gridx = np.zeros(len(xax))
            gridy = np.zeros(len(yax))
            modx = self.frequency%(self.fitbg['results'][self.name]['dnu'][0])
            starty = min(self.frequency)
            for ii in range(len(gridx)):
                for jj in range(len(gridy)):
                    if self.bg_corr[((modx >= startx)&(modx < startx+(self.fitbg['results'][self.name]['dnu'][0]/nox)))&((self.frequency >= starty)&(self.frequency < starty+self.fitbg['results'][self.name]['dnu'][0]))] != []:
                        arr[ii, jj] = np.sum(self.bg_corr[((modx >= startx)&(modx < startx+(self.fitbg['results'][self.name]['dnu'][0]/n_across)))&((self.frequency >= starty)&(self.frequency < starty+self.fitbg['results'][self.name]['dnu'][0]))])
                    else:
                        arr[ii, jj] = np.nan
                    gridy[jj] = starty + (self.fitbg['results'][self.name]['dnu'][0]/2.0)
                    starty += self.fitbg['results'][self.name]['dnu'][0]
                gridx[ii] = startx + (self.fitbg['results'][self.name]['dnu'][0]/self.globe['n_across']/2.0)
                starty = min(self.frequency)
                startx += self.fitbg['results'][self.name]['dnu'][0]/self.globe['n_across']
            smoothed = arr
            dim = smoothed.shape

            smoothed_2 = np.zeros((2*dim[0], dim[1]))
            smoothed_2[0:dim[0],:] = smoothed
            smoothed_2[dim[0]:(2*dim[0]),:] = smoothed
            smoothed = np.swapaxes(smoothed_2, 0, 1)
            extent = [min(gridx)-(self.fitbg['results'][self.name]['dnu'][0]/n_across/2.0), 2*max(gridx)+(self.fitbg['results'][self.name]['dnu'][0]/n_across/2.0), min(gridy)-(self.fitbg['results'][self.name]['dnu'][0]/2.0), max(gridy)+(self.fitbg['results'][self.name]['dnu'][0]/2.0)]
            gridx = np.array(list(gridx)+list(gridx+self.fitbg['results'][self.name]['dnu'][0]))
            return smoothed, gridx, gridy, extent


    def get_red_noise(self):
        """
        Calculates red noise level, or stellar background contribution, from power spectrum.

        Returns
        -------
        result : bool
            will return `False` if model converges, otherwise `True`.
        
        """
        constants = utils.Constants()
        # Use as initial guesses for the optimized model
        try:
            pars, _ = curve_fit(self.fitbg['functions'][self.nlaws], self.bin_freq, self.bin_pow, p0=self.pars, sigma=self.bin_err, bounds=self.bounds)
        except RuntimeError as _:
            return True
        else:
            self.pars = pars
            self.bg_corr = self.random_pow/models.harvey(self.frequency, self.pars, total=True)
            # save final values for Harvey components
            for n in range(self.nlaws):
                self.fitbg['results'][self.name]['tau_%d'%(n+1)].append(self.pars[2*n]*10**6.)
                self.fitbg['results'][self.name]['sigma_%d'%(n+1)].append(self.pars[2*n+1])
            self.fitbg['results'][self.name]['white'].append(self.pars[2*self.nlaws])
        return False


    def estimate_dnu(self):
        """
        Estimate a value for $\delta \nu$.

        Returns
        -------
        None
        
        """
        # define the peak in the ACF
        zoom_lag = self.lag[(self.lag>=self.fitbg['acf_mask'][self.name][0])&(self.lag<=self.fitbg['acf_mask'][self.name][1])]
        zoom_auto = self.auto[(self.lag>=self.fitbg['acf_mask'][self.name][0])&(self.lag<=self.fitbg['acf_mask'][self.name][1])]

		      # fit a Gaussian function to the selected peak in the ACF
        p_gauss3, _ = curve_fit(models.gaussian, zoom_lag, zoom_auto, p0=self.acf_guesses, bounds=self.acf_bb[0])
        # the center of that Gaussian is our estimate for Dnu
        dnu = p_gauss3[2]
        self.fitbg['results'][self.name]['dnu'].append(dnu)


import os
import glob
import numpy as np
import pandas as pd
from tqdm import tqdm
from itertools import chain
from astropy.io import ascii
import multiprocessing as mp
from astropy.stats import mad_std
from astropy.timeseries import LombScargle as lomb

from pysyd.plots import set_plot_params
from pysyd.functions import *
from pysyd.models import *



def get_info_all(args, parallel, CLI=True):
    """
    Loads todo.txt, sets up file paths, loads in any available star information, saves the 
    relevant parameters for each of the two main routines and sets the plotting parameters.

    Parameters
    ----------
    args : argparse.Namespace
        command-line arguments
    parallel : bool
        if pysyd will be running in parallel mode
    CLI : bool, optional
        if CLI is not being used (i.e. `False`), the modules draw default values from a different location

    Returns
    -------
    args : argparse.Namespace
        the updated command-line arguments

    """
    # Get parameters for all modules
    args = get_pysyd_parameters(args, parallel, CLI)
    # Get invidual/specific star info from csv file (if it exists)
    args = get_csv_info(args, CLI)
    if CLI:
        # Check the input variables
        check_input_args(args)
        args = get_command_line(args)
    set_plot_params()

    return args


def get_pysyd_parameters(args, parallel, CLI):
    """
    Basic function to call the individual functions that load and
    save parameters for different modules.

    Parameters
    ----------
    args : argparse.Namespace
        command-line arguments
    CLI : bool
        `True` if running pysyd via command line

    Returns
    -------
    args : argparse.Namespace
        the updated command-line arguments

    """

    # Initialize main 'params' dictionary
    args = get_main_params(args, CLI)
    args = get_groups(args, parallel)
    # Initialize parameters for the find excess routine
    args = get_excess_params(args, CLI)
    # Initialize parameters for the fit background routine
    args = get_background_params(args, CLI)
    # Initialize parameters relevant for estimating global parameters
    args = get_global_params(args, CLI)

    return args


def get_main_params(args, CLI, stars=None, verbose=False, show=False, save=True, 
                    kepcorr=False, of_actual=None, of_new=None,):
    """
    Get the parameters for the find excess routine.

    Parameters
    ----------
    args : argparse.Namespace
        the command line arguments
    stars : List[str], optional
        list of targets to process. If `None`, will read in from `info/todo.txt` (default).
    verbose : bool, optional
        turn on verbose output. Default is `False`.
    show : bool, optional
        show output figures. Default is `False`.
    save : bool, optional
        save all data products. Default is `True`.
    kepcorr : bool, optional
        use the module that corrects for known kepler artefacts. Default is `False`.
    of_actual : int, optional
        oversampling factor of input PS. Default value is `None`.
    of_new : int, optional
        oversampling factor of newly-computed PS. Default value is `None`.

    Returns
    -------
    args : argparse.Namespace
        the updated command line arguments
    args.params : Dict[str,object]
        the parameters of higher-level functionality

    """
    if CLI:
        params = {
            'stars': args.stars,
            'inpdir': args.inpdir,
            'outdir': args.outdir,
            'info': args.info,
            'show': args.show,
            'save': args.save,
            'of_actual': args.of_actual,
            'of_new': args.of_new,
            'kepcorr': args.kepcorr,
        }
    else:
        args.todo = os.path.join(os.path.abspath(os.getcwd()), 'info', 'todo.txt')
        args.info = os.path.join(os.path.abspath(os.getcwd()), 'info', 'star_info.csv')
        args.inpdir = os.path.join(os.path.abspath(os.getcwd()), 'data')
        args.of_actual = of_actual
        args.of_new = of_new
        args.kepcorr = kepcorr
        params = {
            'stars': stars,
            'inpdir': os.path.join(os.path.abspath(os.getcwd()), 'data'),
            'outdir': os.path.join(os.path.abspath(os.getcwd()), 'results'),
            'info': args.info,
            'show': show,
            'save': save,
        }

    # Open star list
    if params['stars'] is None or params['stars'] == []:
        with open(args.todo, "r") as f:
            params['stars'] = [line.strip().split()[0] for line in f.readlines()]

    # Set file paths and make directories if they don't yet exist
    for star in params['stars']:
        params[star] = {}
        params[star]['path'] = os.path.join(params['outdir'], star)
        if params['save'] and not os.path.exists(params[star]['path']):
            os.makedirs(params[star]['path'])
        params[star]['ech_mask'] = None
    args.params = params

    return args


def get_groups(args, parallel=False):
    """
    Sets up star groups to run in parallel based on the number of threads.

    Parameters
    ----------
    args : argparse.Namespace
        command line arguments
    parallel : bool
        run pySYD in parallel

    Returns
    -------
    args : argparse.Namespace
        the updated command line arguments
    args.params['groups'] : ndarray
        star groups to process (groups == number of threads)

    Returns
    ----------
    None

    """
    if parallel:
        todo = np.array(args.params['stars'])
        args.params['verbose'] = False
        args.params['show'] = False
        if args.n_threads == 0:
            args.n_threads = mp.cpu_count()
        if len(todo) < args.n_threads:
            args.n_threads = len(todo)
        # divide stars into groups set by number of cpus/nthreads available
        digitized = np.digitize(np.arange(len(todo))%args.n_threads,np.arange(args.n_threads))
        args.params['groups'] = np.array([todo[digitized == i] for i in range(1, args.n_threads+1)], dtype=object)
    else:
        args.params['groups'] = np.array([])

    return args


def get_excess_params(args, CLI, n_trials=3, step=0.25, binning=0.005, smooth_width=50.0, 
                      mode='mean'):
    """
    Get the parameters for the find excess routine.

    Parameters
    ----------
    args : argparse.Namespace
        the command line arguments
    step : float, optional
        TODO: Write description. Default value is `0.25`.
    binning : float, optional
        logarithmic binning width. Default value is `0.005`.
    mode : {'mean', 'median', 'gaussian'}
        mode to use when binning
    n_trials : int, optional
        the number of trials. Default value is `3`.

    Returns
    -------
    args : argparse.Namespace
        the updated command line arguments
    args.findex : Dict[str,object]
        the parameters of the find excess routine

    """
    if CLI:
        findex = {
            'step': args.step,
            'binning': args.binning,
            'mode': args.mode,
            'smooth_width': args.smooth_width,
            'n_trials': args.n_trials,
            'results': {},
        }
    else:
        findex = {
            'step': step,
            'binning': binning,
            'mode': mode,
            'smooth_width': smooth_width,
            'n_trials': n_trials,
            'results': {},
        }
    args.findex = findex

    return args


def get_background_params(args, CLI, ind_width=20.0, box_filter=1.0, n_rms=20, mc_iter=1, 
                          samples=False, n_laws=None, ab=False, use='bic',):
    """
    Get the parameters for the background-fitting routine.

    Parameters
    ----------
    args : argparse.Namespace
        the command line arguments
    box_filter : float
        the size of the 1D box smoothing filter (in muHz). Default value is `1.0`.
    ind_width : float
        the independent average smoothing width (in muHz). Default value is `20.0`.
    n_rms : int
        number of data points to estimate red noise contributions. Default value is `20`.
    use : str
        which metric to use (i.e. bic or aic) for model selection. Default is `'bic'`.
    ab : bool
        use {a,b} parametrization for Harvey models. Default is `False`.
    n_laws : int
        force number of Harvey-like components in background fit. Default value is `None`.
    mc_iter : int
        number of samples used to estimate uncertainty. Default value is `1`.
    samples : bool
        if true, will save the monte carlo samples to a csv. Default value is `False`.
    args.n_peaks : int
        the number of peaks to select. Default value is `10`.
    args.force : float
        if not false (i.e. non-zero) will force dnu to be the equal to this value. 
    args.clip : bool
        if true will set the minimum frequency value of the echelle plot to `clip_value`. Default value is `True`.
    args.clip_value : float
        the minimum frequency of the echelle plot. Default value is `0.0`.
    args.smooth_ech : float
        option to smooth the output of the echelle plot
    args.smooth_ps : float
        frequency with which to smooth power spectrum. Default value is `1.0`.
    args.slope : bool
        if true will correct for edge effects and residual slope in Gaussian fit. Default value is `False`.
    args.samples : bool
        if true, will save the monte carlo samples to a csv. Default value is `True`.
    args.convert : bool
        converts Harvey parametrization to physical quantities {a_n,b_n} -> {tau_n,sigma_n}. Default value is `True`.
    args.drop : bool
        drops the extra columns after converting the samples. Default value is `True`.

    Returns
    -------
    args : argparse.Namespace
        the updated command line arguments
    args.fitbg : Dict[str,object]
        the parameters relevant for the fit background routine

    """
    if CLI:
        fitbg = {
            'ind_width': args.ind_width,
            'box_filter': args.box_filter,
            'n_rms': args.n_rms,
            'n_laws': args.n_laws,
            'ab': args.ab,
            'metric': args.use,
            'functions': {0: harvey_none, 1: harvey_one, 2: harvey_two, 3: harvey_three},
            'mc_iter': args.mc_iter,
            'samples': args.samples,
            'results': {},
        }
    else:
        fitbg = {
            'ind_width': ind_width,
            'box_filter': box_filter,
            'n_rms': n_rms,
            'n_laws': n_laws,
            'ab': args.ab,
            'metric': use,
            'functions': {0: harvey_none, 1: harvey_one, 2: harvey_two, 3: harvey_three},
            'mc_iter': mc_iter,
            'samples': samples,
            'results': {},
        }
    args.fitbg = fitbg

    return args


def get_global_params(args, CLI, sm_par=None, numax=None, lower_ps=None, upper_ps=None,
                      width=1.0, dnu=None, smooth_ps=2.5, threshold=1.0, n_peaks=5, 
                      clip_ech=True, clip_value=None, smooth_ech=None, interp_ech=False, 
                      lower_ech=None, upper_ech=None, n_across=50, n_down=5,):
    """
    Get the parameters relevant for finding global asteroseismic parameters numax and dnu.

    Parameters
    ----------
    args : argparse.Namespace
        the command line arguments
    sm_par : float
        Gaussian filter width for determining smoothed numax (values are typically between 1-4)
    numax : float
        guess for numax
    lower_ps : float
        lower bound of power excess (in muHz). Default value is `None`.
    upper_ps : float
        upper bound of power excess (in muHz). Default value is `None`.
    width : float
        fractional width to use for power excess centerd on numax. Default value is `1.0`.
    smooth_ps : float
        box filter [in muHz] for PS smoothing before calculating ACF. Default value is `1.5`.
    threshold : float
        fractional width of FWHM to use in ACF for later iterations. Default value is `1.0`.
    n_peaks : int
        the number of peaks to select. Default value is `10`.
    lower_ech : float
        lower bound of folded PS (in muHz) to 'whiten' mixed modes. Default value is `None`.
    upper_ech : float
        upper bound of folded PS (in muHz) to 'whiten' mixed modes. Default value is `None`.
    clip : bool
        if true will set the minimum frequency value of the echelle plot to `clip_value`. Default value is `True`.
    clip_value : float
        the minimum frequency of the echelle plot. Default value is `0.0`.
    smooth_ech : float
        option to smooth the output of the echelle plot
    interp_ech : bool
        turns on the bilinear smoothing in echelle plot
    n_across : int
        x-axis resolution on the echelle diagram. Default value is `50`. (NOT CURRENTLY IMPLEMENTED YET)
    n_down : int
        how many radial orders to plot on the echelle diagram. Default value is `5`. (NOT CURRENTLY IMPLEMENTED YET)

    Returns
    -------
    args : argparse.Namespace
        the updated command line arguments
    args.globe : Dict[str,object]
        the parameters relevant for determining the global parameters routine

    """
    if CLI:
        globe = {
            'sm_par': args.sm_par,
            'width': args.width,
            'smooth_ps': args.smooth_ps,
            'threshold': args.threshold,
            'n_peaks': args.n_peaks,
            'clip_ech': args.clip_ech,
            'clip_value': args.clip_value,
            'smooth_ech': args.smooth_ech,
            'interp_ech': args.interp_ech,
            'n_across': args.n_across,
            'n_down': args.n_down,
        }
    else:
        globe = {
            'sm_par': sm_par,
            'width': width,
            'smooth_ps': smooth_ps,
            'threshold': threshold,
            'n_peaks': n_peaks,
            'clip_ech': clip_ech,
            'clip_value': clip_value,
            'smooth_ech': smooth_ech,
            'interp_ech': interp_ech,
            'n_across': n_across,
            'n_down': n_down,
        }
    args.globe = globe

    return args


def get_csv_info(args, CLI=True, run_excess=True):
    """
    Reads in any star information provided via args.info and is 'info/star_info.csv' by default. 
    ** Please note that this is NOT required for pySYD to run successfully **

    Parameters
    ----------
    args : argparse.Namespace
        the command line arguments
    columns : list
        the list of columns to provide stellar information for

    Returns
    -------
    args : argparse.Namespace
        the updated command line arguments

    """
    constants = Constants()
    columns = get_data_columns(type='required')
    # Open file if it exists
    if os.path.exists(args.info):
        df = pd.read_csv(args.info)
        stars = [str(each) for each in df.stars.values.tolist()]
        for i, star in enumerate(args.params['stars']):
            args.params[star]['excess'] = run_excess
            args.params[star]['force'] = False
            if star in stars:
                idx = stars.index(star)
                # Update information from columns
                for column in columns:
                    if not np.isnan(float(df.loc[idx,column])):
                        args.params[star][column] = float(df.loc[idx, column])
                    else:
                        args.params[star][column] = None
                # Add estimate of numax if the column exists
                if args.params[star]['numax'] is not None:
                    args.params[star]['dnu'] = 0.22*(args.params[star]['numax']**0.797)
                # Otherwise estimate using other stellar parameters
                else:
                    if args.params[star]['radius'] is not None and args.params[star]['logg'] is not None:
                        args.params[star]['mass'] = ((((args.params[star]['radius']*constants.r_sun)**(2.0))*10**(args.params[star]['logg'])/constants.G)/constants.m_sun)
                        args.params[star]['numax'] = constants.numax_sun*args.params[star]['mass']*(args.params[star]['radius']**(-2.0))*((args.params[star]['teff']/constants.teff_sun)**(-0.5))
                        args.params[star]['dnu'] = constants.dnu_sun*(args.params[star]['mass']**(0.5))*(args.params[star]['radius']**(-1.5))  
            # if target isn't in csv, still save basic parameters called througout the code
            else:
                for column in columns:
                    args.params[star][column] = None
    # same if the file does not exist
    else:
        for star in args.stars:
            args.params[star]['excess'] = run_excess
            args.params[star]['force'] = False
            for column in columns:
                args.params[star][column] = None
    return args


def check_input_args(args, max_laws=3):
    """ 
    Make sure that any command-line inputs are the proper lengths, types, etc.

    Parameters
    ----------
    args : argparse.Namespace
        the command line arguments
    max_laws : int
        maximum number of resolvable Harvey components

    Yields
    ------
    ???

    """

    checks={'lower_bg':args.lower_bg,'upper_bg':args.upper_bg,'lower_ex':args.lower_ex,
            'upper_ex':args.upper_ex,'lower_ps':args.lower_ps,'upper_ps':args.upper_ps,
            'lower_ech':args.lower_ech,'upper_ech':args.upper_ech,'dnu':args.dnu,'numax':args.numax}
    for check in checks:
        if checks[check] is not None:
            assert len(args.stars) == len(checks[check]), "The number of values provided for %s does not equal the number of stars"%check
    if args.of_actual is not None:
        assert isinstance(args.of_actual,int), "The oversampling factor for the input PS must be an integer"
    if args.of_new is not None:
        assert isinstance(args.of_new,int), "The new oversampling factor must be an integer"
    if args.n_laws is not None:
        assert args.n_laws <= max_laws, "We likely cannot resolve %d Harvey-like components for point sources. Please select a smaller number."%args.n_laws


def get_command_line(args, numax=None, dnu=None, lower_ps=None, upper_ps=None, 
                     lower_ech=None, upper_ech=None):
    """
    If certain CLI options are provided, it saves it to the appropriate star. This
    is called after the csv is checked and therefore, this will override any duplicated
    information provided there (if applicable).

    Parameters
    ----------
    args : argparse.Namespace
        the command line arguments
    args.lower_ps : float, optional
        the lower frequency bound for numax (in muHz). Default is `None`.
    args.upper_ps : float, optional
        the upper frequency bound for numax (in muHz). Default is `None`.
    args.numax : List[float], optional
        the estimated numax (in muHz). Default is `None`.
    args.dnu : List[float], optional
        the estimated frequency spacing or dnu (in muHz). Default is `None`.
    args.lower_ech : List[float], optional
        the lower frequency for whitening the folded PS (in muHz). Default is `None`.
    args.upper_ech : List[float], optional
        the upper frequency for whitening the folded PS (in muHz). Default is `None`.

    Returns
    -------
    args : argparse.Namespace
        the updated command line arguments

    """

    override = {
        'lower_bg': args.lower_bg,
        'upper_bg': args.upper_bg,
        'lower_ex': args.lower_ex,
        'upper_ex': args.upper_ex,
        'lower_ps': args.lower_ps,
        'upper_ps': args.upper_ps,
        'numax': args.numax,
        'dnu': args.dnu,
        'lower_ech': args.lower_ech,
        'upper_ech': args.upper_ech,
    }

    for i, star in enumerate(args.params['stars']):
        for each in override:
            if override[each] is not None:
                # if numax is provided via CLI, findex is skipped
                if each == 'numax':
                    args.params[star]['excess'] = False
                    args.params[star]['numax'] = override[each][i]
                    args.params[star]['dnu'] = 0.22*(args.params[star]['numax']**0.797)
                # if dnu is provided via CLI, this value is used instead of the derived dnu
                elif each == 'dnu':
                    args.params[star]['force'] = True
                    args.params[star]['guess'] = override[each][i]
                else:
                    args.params[star][each] = override[each][i]
        if args.params[star]['force']:
            args.params[star]['dnu'] = args.params[star]['guess']
        if args.params[star]['lower_ech'] is not None and args.params[star]['upper_ech'] is not None:
            args.params[star]['ech_mask'] = [args.params[star]['lower_ech'],args.params[star]['upper_ech']]
        else:
            args.params[star]['ech_mask'] = None

    return args


def load_data(star, args):
    """
    Loads both the light curve and power spectrum data in for a given star,
    which will return `False` if unsuccessful and therefore, not run the rest
    of the pipeline.

    Parameters
    ----------
    star : target.Target
        the pySYD pipeline object
    args : argparse.Namespace
        command line arguments

    Returns
    -------
    star : target.Target
        the pySYD pipeline object
    star.lc : bool
        will return `True` if the light curve data was loaded in properly otherwise `False`
    star.ps : bool
        will return `True` if the power spectrum file was successfully loaded otherwise `False`

    """
    star.pickles=[]
    # Now done at beginning to make sure it only does this once per star
    if glob.glob(os.path.join(args.inpdir,'%s*'%str(star.name))) != []:
        if star.verbose:
            print('\n\n------------------------------------------------------')
            print('Target: %s'%str(star.name))
            print('------------------------------------------------------')

        # Load light curve
        args, star, note = load_time_series(args, star)
        if star.verbose:
            print(note)

        # Load power spectrum
        args, star, note = load_power_spectrum(args, star)
        if star.verbose:
            print(note)

    return star


def load_file(path):
    """
    Load a light curve or a power spectrum from a basic 2xN txt file
    and stores the data into the `x` (independent variable) and `y`
    (dependent variable) arrays, where N is the length of the series.

    Parameters
    ----------
    path : str
        the file path of the data file

    Returns
    -------
    x : numpy.array
        the independent variable i.e. the time or frequency array 
    y : numpy.array
        the dependent variable, in this case either the flux or power array

    """
    f = open(path, "r")
    lines = f.readlines()
    f.close()
    # Set values
    x = np.array([float(line.strip().split()[0]) for line in lines])
    y = np.array([float(line.strip().split()[1]) for line in lines])
    return x, y


def load_time_series(args, star, note=''):
    """
    If available, star.lc is set to `True`, the time series data
    is loaded in, and then it calculates the cadence and nyquist 
    freqency. If time series data is not provided, either the
    cadence or nyquist frequency must be provided via CLI

    Parameters
    ----------
    star : target.Target
        the pySYD pipeline object
    args : argparse.Namespace
        command line arguments
    args.cadence : int
        cadence of time series data (if known but data is not available)
    args.nyquist : float
        nyquist frequency of the provided power spectrum
    note : str
        optional suppressed verbose output

    Returns
    -------
    star : target.Target
        the pySYD pipeline object
    star.lc : bool
        will return `True` if the light curve data was loaded in properly otherwise `False`
    star.time : numpy.array
        time array in days
    star.flux : numpy.array
        relative or normalized flux array

    """
    star.lc = False
    star.nyquist = None
    # Try loading the light curve
    if os.path.exists(os.path.join(args.inpdir, '%s_LC.txt'%star.name)):
        star.lc = True
        star.time, star.flux = load_file(os.path.join(args.inpdir, '%s_LC.txt'%star.name))
        star.time -= min(star.time)
        star.cadence = int(round(np.nanmedian(np.diff(star.time)*24.0*60.0*60.0),0))
        star.nyquist = 10**6./(2.0*star.cadence)
        star.baseline = (max(star.time)-min(star.time))*24.*60.*60.
        note += '# LIGHT CURVE: %d lines of data read\n# Time series cadence: %d seconds'%(len(star.time),star.cadence)

    return args, star, note


def load_power_spectrum(args, star, note='', long=10**6):
    """
    Loads in the power spectrum data in for a given star,
    which will return `False` if unsuccessful and therefore, not run the rest
    of the pipeline.

    Parameters
    ----------
    star : target.Target
        the pySYD pipeline object
    args : argparse.Namespace
        command line arguments
    args.kepcorr : bool
        if true, will run the module to mitigate the Kepler artefacts in the power spectrum. Default is `False`.
    args.of_actual : int
        the oversampling factor, if the power spectrum is already oversampled. Default is `1`, assuming a critically sampled PS.
    args.of_new : float
        the oversampling factor to use for the first iterations. Default is `5`.
    note : str
        optional suppressed verbose output
    long : int
        will display a warning if length of PS is longer than 10**6 lines 

    Returns
    -------
    star : target.Target
        the pySYD pipeline object
    star.ps : bool
        will return `True` if the power spectrum file was successfully loaded otherwise `False`
    star.frequency : numpy.array
        frequency array in muHz
    star.power : numpy.array
        power spectral density array

    """
    star.ps = False
    # Try loading the power spectrum
    if not os.path.exists(os.path.join(args.inpdir, '%s_PS.txt'%star.name)):
        note += '# ERROR: %s/%s_PS.txt not found\n'%(args.inpdir, star.name)
    else:
        star.ps = True
        star.frequency, star.power = load_file(os.path.join(args.inpdir, '%s_PS.txt'%star.name))
        note += '# POWER SPECTRUM: %d lines of data read\n'%len(star.frequency)
        if len(star.frequency) >= long:
            note += '# WARNING: PS is large and will slow down the software'
        star.resolution = star.frequency[1]-star.frequency[0]
        if args.kepcorr:
            note += '# **using Kepler artefact correction**\n'
            star = remove_artefact(star)
        if star.params[star.name]['ech_mask'] is not None:
            note += '# **whitening the PS to remove mixed modes**\n'
            star = whiten_mixed(star)
        args, star, note = check_input(args, star, note)

    return args, star, note


def check_input(args, star, note):
    """
    Checks the type(s) of input data and creates any additional, optional
    arrays as well as critically-sampled power spectra (when applicable).

    Parameters
    ----------
    args : argparse.Namespace
        command line arguments
    star : target.Target
        pySYD target object
    note : str, optional
        optional verbose output

    Returns
    -------
    args : argparse.Namespace
        updated command line arguments
    star : target.Target
        updated pySYD target object
    note : str, optional
        updated optional verbose output

    """
    if star.lc:
        args.of_actual = int(round((1./((max(star.time)-min(star.time))*0.0864))/(star.frequency[1]-star.frequency[0])))
        star.freq_cs = np.array(star.frequency[args.of_actual-1::args.of_actual])
        star.pow_cs = np.array(star.power[args.of_actual-1::args.of_actual])
        if args.of_new is not None:
            note += '# Computing new PS using oversampling of %d\n'%args.of_new
            freq_os, pow_os = lomb(star.time, star.flux).autopower(method='fast', samples_per_peak=args.of_new, maximum_frequency=star.nyquist)
            star.freq_os = freq_os*(10.**6/(24.*60.*60.))
            star.pow_os = 4.*pow_os*np.var(star.flux*1e6)/(np.sum(pow_os)*(star.freq_os[1]-star.freq_os[0]))
        else:
            star.freq_os, star.pow_os = np.copy(star.frequency), np.copy(star.power)
    else:
        if args.of_actual is not None:
            star.freq_cs = np.array(star.frequency[args.of_actual-1::args.of_actual])
            star.pow_cs = np.array(star.power[args.of_actual-1::args.of_actual])
            star.freq_os, star.pow_os = np.copy(star.frequency), np.copy(star.power)                    
        else:
            star.freq_cs, star.pow_cs = np.copy(star.frequency), np.copy(star.power)
            star.freq_os, star.pow_os = np.copy(star.frequency), np.copy(star.power)
            note += '# WARNING: using input PS with no additional information\n'
            if args.mc_iter > 1:
                note += '# **uncertainties may not be reliable unless using a critically-sampled PS**'
        star.baseline = 1./((star.freq_cs[1]-star.freq_cs[0])*10**-6.)
    if args.of_actual is not None and args.of_actual != 1:
        note += '# PS is oversampled by a factor of %d\n'%args.of_actual
    else:
        note += '# PS is critically-sampled\n'
    note += '# PS resolution: %.6f muHz'%(star.freq_cs[1]-star.freq_cs[0])
    return args, star, note


def get_findex(star):
    """
    Before running the first module (find excess), this masks out any unwanted
    frequency regions and also gets appropriate bin sizes for the collapsed ACF
    function if some prior information on numax is provided.

    Parameters
    ----------
    star : target.Target
        pySYD target object

    Returns
    -------
    star : target.Target
        updated pySYD target object

    """
    # If running the first module, mask out any unwanted frequency regions
    star.frequency, star.power = np.copy(star.freq_os), np.copy(star.pow_os)
    star.resolution = star.frequency[1]-star.frequency[0]
    # mask out any unwanted frequencies
    if star.params[star.name]['lower_ex'] is not None:
        lower = star.params[star.name]['lower_ex']
    else:
        lower = min(star.frequency)
    if star.params[star.name]['upper_ex'] is not None:
        upper = star.params[star.name]['upper_ex']
    else:
        upper = max(star.frequency)
        if star.nyquist is not None and star.nyquist < upper:
            upper = star.nyquist
    star.freq = star.frequency[(star.frequency >= lower)&(star.frequency <= upper)]
    star.pow = star.power[(star.frequency >= lower)&(star.frequency <= upper)]
    if (star.params[star.name]['numax'] is not None and star.params[star.name]['numax'] <= 500.) or (star.nyquist is not None and star.nyquist <= 300.):
        star.boxes = np.logspace(np.log10(0.5), np.log10(25.), star.findex['n_trials'])*1.
    else:
        star.boxes = np.logspace(np.log10(50.), np.log10(500.), star.findex['n_trials'])*1.
    return star


def check_fitbg(star):
    """
    Checks if there is a starting value for numax as pySYD needs this information to begin the 
    second module (whether be it from the first module, CLI or saved to info/star_info.csv).

    Returns
    -------
    result : bool
        will return `True` if there is prior value for numax otherwise `False`.

    """
    # Check if numax was provided as input
    if star.params[star.name]['numax'] is None:
        # If not, checks if findex was run
        if glob.glob(os.path.join(star.params[star.name]['path'],'excess.csv')) != []:
            df = pd.read_csv(os.path.join(star.params[star.name]['path'],'excess.csv'))
            for col in ['numax', 'dnu', 'snr']:
                star.params[star.name][col] = df.loc[0, col]
        # Break if no numax is provided in any scenario
        else:
            print('# ERROR: the second module cannot run without any prior info for numax')
            return False
    return True


def get_fitbg(star):
    """
    Gets initial guesses for granulation components (i.e. timescales and amplitudes) using
    solar scaling relations. This resets the power spectrum and has its own independent
    filter (i.e. [lower,upper] mask) to use for this subroutine.

    Parameters
    ----------
    star : target.Target
        pySYD target object
    star.oversample : bool
        if `True`, it will use an oversampled power spectrum for the first iteration or 'step'
    minimum_freq : float
        minimum frequency to use for the power spectrum if `None` is provided (via info/star_info.csv). Default = `10.0` muHz. Please note: this is typically sufficient for most stars but may affect evolved stars!
    maximum_freq : float
        maximum frequency to use for the power spectrum if `None` is provided (via info/star_info.csv). Default = `5000.0` muHz.

    Returns
    -------
    star : target.Target
        updated pySYD target object

    """
    star.frequency, star.power = np.copy(star.freq_os), np.copy(star.pow_os)
    star.resolution = star.frequency[1]-star.frequency[0]

    if star.params[star.name]['lower_bg'] is not None:
        lower = star.params[star.name]['lower_bg']
    else:
        lower = min(star.frequency)
    if star.params[star.name]['upper_bg'] is not None:
        upper = star.params[star.name]['upper_bg']
    else:
        upper = max(star.frequency)
        if star.nyquist is not None and star.nyquist < upper:
            upper = star.nyquist
    star.params[star.name]['bg_mask']=[lower,upper]

    # Mask power spectrum for fitbg module
    mask = np.ma.getmask(np.ma.masked_inside(star.frequency, star.params[star.name]['bg_mask'][0], star.params[star.name]['bg_mask'][1]))
    star.frequency, star.power = np.copy(star.frequency[mask]), np.copy(star.power[mask])
    star.random_pow = np.copy(star.power)
    # Get other relevant initial conditions
    star.i = 0
    star.fitbg['results'][star.name] = {'numax_smooth':[],'A_smooth':[],'numax_gauss':[],'A_gauss':[],
                                        'FWHM':[],'dnu':[],'white':[]}

    # Use scaling relations from sun to get starting points
    star = solar_scaling(star)

    # if lower numax adjust default smoothing filter from 2.5->0.5muHz
    # this needs to be fixed - it doesn't change for the rest
    if star.params[star.name]['numax'] <= 500.:
        star.fitbg['ind_width'] = 5.0
        star.globe['smooth_ps'] = 0.5
    else:
        star.fitbg['ind_width'] = 20.0
        star.globe['smooth_ps'] = 2.5

    return star


def solar_scaling(star, scaling='tau_sun_single', max_laws=3, times=1.5):
    """
    Uses scaling relations from the Sun to:
    1) estimate the width of the region of oscillations using numax
    2) guess starting values for granulation timescales

    Parameters
    ----------
    max_laws : int
        the maximum number of resolvable Harvey-like components

    """
    constants = Constants()
    # Use scaling relations to estimate width of oscillation region to mask out of the background fit
    width = constants.width_sun*(star.params[star.name]['numax']/constants.numax_sun)
    maxpower = [star.params[star.name]['numax']-(width*star.globe['width']), 
                star.params[star.name]['numax']+(width*star.globe['width'])]
    if star.params[star.name]['lower_ps'] is not None:
        maxpower[0] = star.params[star.name]['lower_ps']
    if star.params[star.name]['upper_ps'] is not None:
        maxpower[1] = star.params[star.name]['upper_ps']
    star.params[star.name]['ps_mask'] = [maxpower[0],maxpower[1]]

    # Use scaling relation for granulation timescales from the sun to get starting points
    scale = constants.numax_sun/star.params[star.name]['numax']
    if scaling == 'tau_sun_single':
        taus = np.array(constants.tau_sun_single)*scale
    else:
        taus = np.array(constants.tau_sun)*scale
    taus = taus[taus <= star.baseline]
    b = taus*10**-6.
    mnu = (1.0/taus)*10**5.
    star.b = b[mnu >= min(star.frequency)]
    star.mnu = mnu[mnu >= min(star.frequency)]
    if len(star.mnu)==0:
        star.b = b[mnu >= 10.] 
        star.mnu = mnu[mnu >= 10.]
    elif len(star.mnu) > max_laws:
        star.b = b[mnu >= min(star.frequency)][-max_laws:]
        star.mnu = mnu[mnu >= min(star.frequency)][-max_laws:]
    else:
        pass
    # Save copies for plotting after the analysis
    star.nlaws = len(star.mnu)
    star.nlaws_orig = len(star.mnu)
    star.mnu_orig = np.copy(star.mnu)
    star.b_orig = np.copy(star.b)

    return star


def save_file(star, formats=[">15.8f", ">18.10e"]):
    """
    Saves the corrected power spectrum, which is computed by subtracting
    the best-fit stellar background model from the power spectrum.

    Parameters
    ----------
    star : target.Target
        the pySYD pipeline target
    formats : List[str]
        2x1 list of formats to save arrays as
    star.params[star.name]['path'] : str
        path to save the background-corrected power spectrum
    star.frequency : ndarray
        frequency array
    star.bg_corr_sub : ndarray
        background-subtracted power spectrum

    """
    
    f_name=os.path.join(star.params[star.name]['path'],'bgcorr_ps.txt')
    with open(f_name, "w") as f:
        for x, y in zip(star.frequency, star.bg_corr):
            values = [x, y]
            text = '{:{}}'*len(values) + '\n'
            fmt = sum(zip(values, formats), ())
            f.write(text.format(*fmt))
    f.close()
    if star.verbose:
        print(' **background-corrected PS saved**')


def save_findex(star):
    """
    Save the results of the find excess routine into the save folder of the current star.

    Parameters
    ----------
    star : target.Target
        pipeline target with the results of the `find_excess` routine

    """
    best = star.findex['results'][star.name]['best']
    variables = ['star', 'numax', 'dnu', 'snr']
    results = [star.name, star.findex['results'][star.name][best]['numax'], star.findex['results'][star.name][best]['dnu'], star.findex['results'][star.name][best]['snr']]
    save_path = os.path.join(star.params[star.name]['path'],'excess.csv')
    ascii.write(np.array(results), save_path, names=variables, delimiter=',', overwrite=True)


def save_fitbg(star):
    """
    Saves the results of the `fit_background` module.

    Parameters
    ----------
    star : target.Target
        pipeline target with the results of the `fit_background` routine

    """
    df = pd.DataFrame(star.fitbg['results'][star.name])
    star.df = df.copy()
    new_df = pd.DataFrame(columns=['parameter', 'value', 'uncertainty'])
    for c, col in enumerate(df.columns.values.tolist()):
        new_df.loc[c, 'parameter'] = col
        new_df.loc[c, 'value'] = df.loc[0,col]
        if star.fitbg['mc_iter'] > 1:
            new_df.loc[c, 'uncertainty'] = mad_std(df[col].values)
        else:
            new_df.loc[c, 'uncertainty'] = '--'
    new_df.to_csv(os.path.join(star.params[star.name]['path'],'background.csv'), index=False)
    if star.fitbg['samples']:
        df.to_csv(os.path.join(star.params[star.name]['path'],'samples.csv'), index=False)


def verbose_output(star):
    """
    Print results of the `fit_background` routine if verbose is `True`.

    """
    note=''
    df = pd.read_csv(os.path.join(star.params[star.name]['path'],'background.csv'))
    params = get_params_dict()
    if star.fitbg['mc_iter'] > 1:
        note+='\nOutput parameters:'
        line='\n%s: %.2f +/- %.2f %s'
        for idx in df.index.values.tolist():
            note+=line%(df.loc[idx,'parameter'],df.loc[idx,'value'],df.loc[idx,'uncertainty'],params[df.loc[idx,'parameter']]['unit'])
    else:
        note+='----------------------------------------------------\nOutput parameters:'
        line='\n%s: %.2f %s'
        for idx in df.index.values.tolist():
            note+=line%(df.loc[idx,'parameter'],df.loc[idx,'value'],params[df.loc[idx,'parameter']]['unit'])
    note+='\n----------------------------------------------------'
    print(note)


def scrape_output(args):
    """
    Grabs each individual star's results and concatenates results into a single csv in info/ for each submodule
    (i.e. excess.csv and background.csv). This is automatically called if the pySYD is successfully executed for 
    at least one star.

    """

    path = '%s/**/'%args.params['outdir']
    # Findex outputs
    files = glob.glob('%s*excess.csv'%path)
    if files != []:
        df = pd.read_csv(files[0])
        for i in range(1,len(files)):
            df_new = pd.read_csv(files[i])
            df = pd.concat([df, df_new])
        df.to_csv(os.path.join(args.params['outdir'],'excess.csv'), index=False)

    # Fitbg outputs
    files = glob.glob('%s*background.csv'%path)
    if files != []:
        df = pd.DataFrame(columns=['star'])
        for i, file in enumerate(files):
	           df_new = pd.read_csv(file)
	           df_new.set_index('parameter',inplace=True,drop=False)
	           df.loc[i,'star']=file.strip().split('/')[-2]
	           new_header_names=[[i,i+'_err'] for i in df_new.index.values.tolist()]
	           new_header_names=list(chain.from_iterable(new_header_names))          
	           for col in new_header_names:
		              if '_err' in col:
			                 df.loc[i,col]=df_new.loc[col[:-4],'uncertainty']
		              else:
			                 df.loc[i,col]=df_new.loc[col,'value']

        df.fillna('--', inplace=True)
        df.to_csv(os.path.join(args.params['outdir'],'background.csv'), index=False)


def make_latex_table(path, lines='', header=True, columns=None, labels=None, formats=None, units=None, 
                     include_units=True, include_uncertainty=True, save=True, verbose=True):

    # Get data
    df = pd.read_csv(path)
    if columns is None:
        columns = get_data_columns(type='table')
    # Get parameter information
    params = get_params_dict()

    # Make header row
    if header:
        line='Star & '
        for column in columns:
            if include_units:
                line+='%s [%s] & '%(params[column]['latex']['label'],params[column]['latex']['unit'])
            else:
                line+='%s & '%params[column]['latex']['label']
        line = line[:-3] + ' \\\ '
        if verbose:
            print(line)
        lines+=line+'\n'

    # Iterate through all targets
    for i in df.index.values.tolist():
        line='%d & '%df.loc[i,'star']
        for column in columns:
            if df.loc[i,column] != '--':
                if include_uncertainty and '%s_err'%column in df.columns.values.tolist():
                    par = '%s $\pm$ %s & '%(params[column]['latex']['format'], params[column]['latex']['format'])
                    line += par%(df.loc[i,column],df.loc[i,'%s_err'%column])
                else:
                    par = '%s & '%params[column]['latex']['format']
                    line += par%df.loc[i,column]
            else:
                line += '-- & '
        line = line[:-3] + ' \\\ '
        if verbose:
            print(line)
        lines+=line+'\n'

    if save:
        if len(path.split('/')) > 1:
            new_fn='%s/%s.txt'%('/'.join(path.split('/')[:-1]),(path.split('/')[-1]).split('.')[0])
        else:
            new_fn='%s.txt'%(path.split('/')[-1]).split('.')[0]
        f = open(new_fn, "w")
        f.write(lines)
        f.close()


def get_params_dict():

    params={
            'numax_smooth':{'unit':'muHz','label':r'$\rm Smooth \,\, \nu_{max} \,\, [\mu Hz]$','latex':{'label':'$\\nu_{\\mathrm{max}}$', 'format':'%.2f', 'unit':'$\\rm \\mu Hz$'}}, 
            'A_smooth':{'unit':'ppm^2/muHz','label':r'$\rm Smooth \,\, A_{max} \,\, [ppm^{2} \mu Hz^{-1}]$','latex':{'label':'$\\rm A_{osc}$', 'format':'%.2f', 'unit':'$\\rm ppm^{2} \\mu Hz^{-1}$'}}, 
            'numax_gauss':{'unit':'muHz','label':r'$\rm Gauss \,\, \nu_{max} \,\, [\mu Hz]$','latex':{'label':'$\\nu_{\\mathrm{max}}$', 'format':'%.2f', 'unit':'$\\rm \\mu Hz$'}}, 
            'A_gauss':{'unit':'ppm^2/muHz','label':r'$\rm Gauss \,\, A_{max} \,\, [ppm^{2} \mu Hz^{-1}]$','latex':{'label':'$\\rm A_{osc}$', 'format':'%.2f', 'unit':'$\\rm ppm^{2} \\mu Hz^{-1}$'}}, 
            'FWHM':{'unit':'muHz','label':r'$\rm Gauss \,\, FWHM \,\, [\mu Hz]$','latex':{'label':'FWHM', 'format':'%.2f', 'unit':'$\\rm \\mu Hz$'}}, 
            'dnu':{'unit':'muHz','label':r'$\rm \Delta\nu \,\, [\mu Hz]$','latex':{'label':'$\\Delta\\nu$', 'format':'%.2f', 'unit':'$\\rm \\mu Hz$'}},
            'white':{'unit':'ppm^2/muHz','label':r'$\rm White \,\, [ppm^{2} \mu Hz^{-1}]$','latex':{'label':'White noise', 'format':'%.2f', 'unit':'$\\rm ppm^{2} \\mu Hz^{-1}$'}}, 
            'a_1':{'unit':'ppm^2/muHz','label':r'$\rm a_{1} \,\, [ppm^{2} \mu Hz^{-1}]$','latex':{'label':'$\\rm a_{1}$', 'format':'%.2f', 'unit':'$\\rm ppm^{2} \\mu Hz^{-1}$'}},
            'b_1':{'unit':'muHz^-1','label':r'$\rm b_{1} \,\, [\mu Hz^{-1}]$','latex':{'label':'$\rm b_{1}$', 'format':'%.2f', 'unit':'$\rm \mu Hz^{-1}$'}}, 
            'tau_1':{'unit':'s','label':r'$\rm \tau_{1} \,\, [s]$','latex':{'label':'$\tau_{1}$', 'format':'%.2f', 'unit':'s'}},
            'sigma_1':{'unit':'ppm','label':r'$\rm \sigma_{1} \,\, [ppm]$','latex':{'label':'$\sigma_{1}$', 'format':'%.2f', 'unit':'ppm'}}, 
            'a_2':{'unit':'ppm^2/muHz','label':r'$\rm a_{2} \,\, [ppm^{2} \mu Hz^{-1}]$','latex':{'label':'$\rm a_{2}$', 'format':'%.2f', 'unit':'$\rm ppm^{2} \mu Hz^{-1}$'}},
            'b_2':{'unit':'muHz^-1','label':r'$\rm b_{2} \,\, [\mu Hz^{-1}]$','latex':{'label':'$\rm b_{2}$', 'format':'%.2f', 'unit':'$\rm \mu Hz^{-1}$'}}, 
            'tau_2':{'unit':'s','label':r'$\rm \tau_{2} \,\, [s]$','latex':{'label':'$\tau_{2}$', 'format':'%.2f', 'unit':'s'}}, 
            'sigma_2':{'unit':'ppm','label':r'$\rm \sigma_{2} \,\, [ppm]$','latex':{'label':'$\sigma_{2}$', 'format':'%.2f', 'unit':'ppm'}},
            'a_3':{'unit':'ppm^2/muHz','label':r'$\rm a_{3} \,\, [ppm^{2} \mu Hz^{-1}]$','latex':{'label':'$\rm a_{3}$', 'format':'%.2f', 'unit':'$\rm ppm^{2} \mu Hz^{-1}$'}},
            'b_3':{'unit':'muHz^-1','label':r'$\rm b_{3} \,\, [\mu Hz^{-1}]$','latex':{'label':'$\rm b_{3}$', 'format':'%.2f', 'unit':'$\rm \mu Hz^{-1}$'}}, 
            'tau_3':{'unit':'s','label':r'$\rm \tau_{3} \,\, [s]$','latex':{'label':'$\tau_{3}$', 'format':'%.2f', 'unit':'s'}}, 
            'sigma_3':{'unit':'ppm','label':r'$\rm \sigma_{3} \,\, [ppm]$','latex':{'label':'$\sigma_{3}$', 'format':'%.2f', 'unit':'ppm'}}, 
            'a_4':{'unit':'ppm^2/muHz','label':r'$\rm a_{4} \,\, [ppm^{2} \mu Hz^{-1}]$','latex':{'label':'$\rm a_{4}$', 'format':'%.2f', 'unit':'$\rm ppm^{2} \mu Hz^{-1}$'}},
            'b_4':{'unit':'muHz^-1','label':r'$\rm b_{4} \,\, [\mu Hz^{-1}]$','latex':{'label':'$\rm b_{4}$', 'format':'%.2f', 'unit':'$\rm \mu Hz^{-1}$'}}, 
            'tau_4':{'unit':'s','label':r'$\rm \tau_{4} \,\, [s]$','latex':{'label':'$\tau_{4}$', 'format':'%.2f', 'unit':'s'}},
            'sigma_4':{'unit':'ppm','label':r'$\rm \sigma_{4} \,\, [ppm]$','latex':{'label':'$\sigma_{4}$', 'format':'%.2f', 'unit':'ppm'}}, 
            'a_5':{'unit':'ppm^2/muHz','label':r'$\rm a_{5} \,\, [ppm^{2} \mu Hz^{-1}]$','latex':{'label':'$\rm a_{5}$', 'format':'%.2f', 'unit':'$\rm ppm^{2} \mu Hz^{-1}$'}},
            'b_5':{'unit':'muHz^-1','label':r'$\rm b_{5} \,\, [\mu Hz^{-1}]$','latex':{'label':'$\rm b_{5}$', 'format':'%.2f', 'unit':'$\rm \mu Hz^{-1}$'}}, 
            'tau_5':{'unit':'s','label':r'$\rm \tau_{5} \,\, [s]$','latex':{'label':'$\tau_{5}$', 'format':'%.2f', 'unit':'s'}}, 
            'sigma_5':{'unit':'ppm','label':r'$\rm \sigma_{5} \,\, [ppm]$','latex':{'label':'$\sigma_{5}$', 'format':'%.2f', 'unit':'ppm'}}, 
            'a_6':{'unit':'ppm^2/muHz','label':r'$\rm a_{6} \,\, [ppm^{2} \mu Hz^{-1}]$','latex':{'label':'$\rm a_{6}$', 'format':'%.2f', 'unit':'$\rm ppm^{2} \mu Hz^{-1}$'}},
            'b_6':{'unit':'muHz^-1','label':r'$\rm b_{6} \,\, [\mu Hz^{-1}]$','latex':{'label':'$\rm b_{6}$', 'format':'%.2f', 'unit':'$\rm \mu Hz^{-1}$'}}, 
            'tau_6':{'unit':'s','label':r'$\rm \tau_{6} \,\, [s]$','latex':{'label':'$\tau_{6}$', 'format':'%.2f', 'unit':'s'}}, 
            'sigma_6':{'unit':'ppm','label':r'$\rm \sigma_{6} \,\, [ppm]$','latex':{'label':'$\sigma_{6}$', 'format':'%.2f', 'unit':'ppm'}},
            }

    return params


def get_data_columns(type=['csv','required','table']):

    if type == 'csv':
        columns = ['stars','radius','radius_err','teff','teff_err','logg','logg_err','lower_ex','upper_ex','lower_bg','upper_bg','lower_ps','upper_ps','lower_ech','upper_ech','numax','dnu','seed']
    elif type == 'required':
        columns = ['radius','logg','teff','numax','lower_ex','upper_ex','lower_bg','upper_bg','lower_ps','upper_ps','lower_ech','upper_ech','seed']
    elif type == 'table':
        columns = ['numax_smooth', 'A_smooth', 'FWHM', 'dnu', 'white']
    else:
        print("Did not understand that input, please specify type. Choose from: ['full','required','table']")
        columns = []

    return columns


def load_single(args, params={}):
    """
    Loads information for a single star.

    Parameters
    ----------
    args : argparse.Namespace
        command line arguments

    Returns
    -------
    args : argparse.Namespace
        the updated command line arguments

    """
    assert len(args.stars) == 1, "You can only load in data for one star at a time."

    args.star = args.stars[0]
    print(args.stars, args.star)
    check_input_args(args)
    params.update({'inpdir':args.inpdir,'outdir':args.outdir,'of_new':args.of_new,'show':args.show,
                   'of_actual':args.of_actual,'info':args.info,'save':args.save,'kepcorr':args.kepcorr,})

    params['path'] = os.path.join(args.outdir, args.star)
    args.params = params
    # Get star info
    args = get_single_info(args)
    return args


def get_single_info(args):
    """
    Reads in any star information provided via args.info and is 'info/star_info.csv' by default. 
    ** Please note that this is NOT required for pySYD to run successfully **

    Parameters
    ----------
    args : argparse.Namespace
        the command line arguments
    columns : list
        the list of columns to provide stellar information for

    Returns
    -------
    args : argparse.Namespace
        the updated command line arguments

    """
    constants = Constants()
    columns = get_data_columns(type='required')
    # Open file if it exists
    if os.path.exists(args.info):
        df = pd.read_csv(args.info)
        stars = [str(each) for each in df.stars.values.tolist()]
        if args.star in stars:
            idx = stars.index(args.star)
            for column in columns:
                if not np.isnan(float(df.loc[idx,column])):
                    args.params[column] = float(df.loc[idx,column])
                else:
                    args.params[column] = None
        else:
            for column in columns:
                args.params[column] = None
        if args.params['numax'] is not None:
            args.params['dnu'] = 0.22*(args.params['numax']**0.797)
        else:
            if args.params['radius'] is not None and args.params['logg'] is not None:
                args.params['mass'] = ((((args.params['radius']*constants.r_sun)**(2.0))*10**(args.params['logg'])/constants.G)/constants.m_sun)
                args.params['numax'] = constants.numax_sun*args.params['mass']*(args.params['radius']**(-2.0))*((args.params['teff']/constants.teff_sun)**(-0.5))
                args.params['dnu'] = constants.dnu_sun*(args.params['mass']**(0.5))*(args.params['radius']**(-1.5))
    override={'lower_bg':args.lower_bg,'upper_bg':args.upper_bg,'lower_ex':args.lower_ex,
              'upper_ex':args.upper_ex,'lower_ps':args.lower_ps,'upper_ps':args.upper_ps,
              'lower_ech':args.lower_ech,'upper_ech':args.upper_ech,'dnu':args.dnu,'numax':args.numax}
    for each in override:
        if override[each] is not None:
            # if numax is provided via CLI, findex is skipped
            if each == 'numax':
                args.params['numax'] = override[each][0]
                args.params['dnu'] = 0.22*(args.params['numax']**0.797)
            # if dnu is provided via CLI, this value is used instead of the derived dnu
            elif each == 'dnu':
                args.params['force'] = True
                args.params['guess'] = override[each][0]
            else:
                args.params[each] = override[each][0]
                
    return args


class Constants:

    def __init__(self):
        """
        UNITS ARE IN THE SUPERIOR CGS 
        COME AT ME

        """
        # Solar values
        self.m_sun = 1.9891e33
        self.r_sun = 6.95508e10
        self.rho_sun = 1.41
        self.teff_sun = 5777.0
        self.logg_sun = 4.4
        self.teffred_sun = 8907.
        self.numax_sun = 3090.0
        self.dnu_sun = 135.1
        self.width_sun = 1300.0 
        self.tau_sun = [5.2e6,1.8e5,1.7e4,2.5e3,280.0,80.0] 
        self.tau_sun_single = [3.8e6,2.5e5,1.5e5,1.0e5,230.,70.]
        # Constants
        self.G = 6.67428e-8
        # Conversions
        self.cm2au = 6.68459e-14
        self.au2cm = 1.496e+13
        self.rad2deg = 180./np.pi
        self.deg2rad = np.pi/180.


import os
import shutil
import unittest
import subprocess
import numpy as np
import pandas as pd
import multiprocessing as mp

import pysyd
from pysyd import utils
from pysyd.target import Target



def run(args=None, star=None, CLI=True, verbose=False, count=0):
    """
    Main script to run the pySYD pipeline

    Parameters
    ----------
    args : argparse.Namespace
        the command line arguments

    """
    if not CLI and star is not None:
        args = utils.Constants()
        args.verbose = verbose
        args.stars = [star]

    args = utils.get_info_all(args, parallel=False, CLI=CLI)
    for star in args.params['stars']:
        single = load(args=args, star=star, CLI=CLI)
        if hasattr(single, 'ps'):
            count+=1
            single.run_syd()
        else:
            print(' - cannot find data for %s'%single.name)

    # check to make sure that at least one star was successful (count == the number of successfully processed stars)   
    if count != 0:
        if args.verbose:
            print(' - combining results into single csv file')
            print('------------------------------------------------------')
            print()
        # Concatenates output into two files
        utils.scrape_output(args)


def load(args=None, star=None, CLI=True, verbose=False):
    """
    A Target class is initialized and processed for each star in the stargroup.

    Parameters
    ----------
    args : argparse.Namespace
        the command line arguments

    Returns
    -------
    single : target.Target
        current data available for the provided target

    """
    if not CLI and star is not None:
        args = utils.Constants()
        args.verbose = verbose
        args.stars = [star]
        args = utils.get_info_all(args, parallel=False, CLI=CLI)

    single = Target(star, args)

    return single


def parallel(args, CLI=True):
    """
    Uses multiprocessing to run the pySYD pipeline in parallel. Stars are assigned
    evenly to `groups`, which is set by the number of threads or CPUs available.
    Stars will then run one-by-one per group

    Parameters
    ----------
    args : argparse.Namespace
        the command line arguments

    """
    args = utils.get_info_all(args, parallel=True, CLI=CLI)

    # create the separate, asyncrhonous (nthread) processes
    pool = mp.Pool(args.n_threads)
    result_objects = [pool.apply_async(pipe, args=(group, args)) for group in args.params['groups']]
    results = [r.get() for r in result_objects]
    pool.close()
    pool.join()    # postpones execution of the next line until all processes finish
    count = np.sum(results)
      
    # check to make sure that at least one star was successful (count == the number of successfully processed stars)   
    if count != 0:
        if args.verbose:
            print('Combining results into single csv file.')
            print()
        # Concatenates output into two files
        utils.scrape_output(args)


def pipe(group, args, count=0):
    """
    A Target class is initialized and processed for each star in the stargroup.

    Parameters
    ----------
    group : 
        list of stars to be processed as a group
    args : argparse.Namespace
        the command line arguments

    Returns
    -------
    count : int
        the number of successful stars processed by pySYD for a given group of stars

    """

    for star in group:
        load(args, star=star, CLI=CLI)
        single = Target(star, args)
        if single.ps:
            count+=1
            single.run_syd()
    return count


def setup(args, note='', raw='https://raw.githubusercontent.com/ashleychontos/pySYD/master/examples/'):
    """
    Running this after installation will create the appropriate directories in the current working
    directory as well as download example data and files to test your pySYD installation

    Parameters
    ----------
    args : argparse.Namespace
        the command line arguments
    note : Optional[str]
        suppressed (optional verbose output
    path_info : str
        path where star list and information are saved. Default is 'info/'
    raw : str
        path to download package data and examples from (source directory)

    """

    print('\n\nDownloading relevant data from source directory:\n')
    # create info directory
    if len(os.path.split(args.todo)) != 1:
        path_info = os.path.split(args.todo)[0]
        if not os.path.exists(path_info):
            os.mkdir(path_info)
            note+=' - created input file directory: %s \n'%path_info
    print(path_info)
    # get example input files
    infile1='%sinfo/todo.txt'%raw
    outfile1=os.path.join(path_info,'info.txt')
    subprocess.call(['curl %s > %s'%(infile1, outfile1)], shell=True)
    infile2='%sinfo/star_info.csv'%raw
    outfile2=os.path.join(path_info,os.path.split(args.info)[-1])
    subprocess.call(['curl %s > %s'%(infile2, outfile2)], shell=True)

    # if not successful, make empty input files for reference
    if not os.path.exists(outfile1):
        f = open(args.todo, "w")
        f.close()
    if not os.path.exists(outfile2):
        df = pd.DataFrame(columns=utils.get_data_columns(type='csv'))
        df.to_csv(args.info, index=False)

    # create data directory
    if not os.path.exists(args.inpdir):
        os.mkdir(args.inpdir)
        note+=' - created data directory at %s \n'%args.inpdir

    # get example data
    for target in ['1435467', '2309595', '11618103']:
        for ext in ['LC', 'PS']:
            infile='%sdata/%s_%s.txt'%(raw, target, ext)
            outfile=os.path.join(args.inpdir, '%s_%s.txt'%(target, ext))
            subprocess.call(['curl %s > %s'%(infile, outfile)], shell=True)
    print('\n')
    note+=' - example data saved\n'

    # create results directory
    if not os.path.exists(args.outdir):
        os.mkdir(args.outdir)
    note+=' - results will be saved to %s \n\n'%args.outdir
    
    if args.verbose:
        print(note)


import argparse

import pysyd
from pysyd import pipeline
from pysyd import TODODIR, INFODIR, INPDIR, OUTDIR



def main():
    # Properties inherent to both modules
    parser = argparse.ArgumentParser(
                                     description="pySYD: Automated Extraction of Global Asteroseismic Parameters", 
                                     prog='pySYD',
    )
    parser.add_argument('-version', '--version',
                        action='version',
                        version="%(prog)s {}".format(pysyd.__version__),
                        help="Print version number and exit."
    )

    # In the parent parser, we define arguments and options common to all subcommands
    parent_parser = argparse.ArgumentParser(add_help=False)
    parent_parser.add_argument('--file', '--list', '--todo',
                               metavar='path',
                               dest='todo',
                               help='List of stars to process',
                               type=str,
                               default=TODODIR,
    )
    parent_parser.add_argument('--in', '--input', '--inpdir', 
                               metavar='path',
                               dest='inpdir',
                               help='Input directory',
                               type=str,
                               default=INPDIR,
    )
    parent_parser.add_argument('--info', '--information',
                               metavar='path',
                               dest='info',
                               help='Path to star info',
                               type=str,
                               default=INFODIR,
    )
    parent_parser.add_argument('--out', '--outdir', '--output',
                               metavar='path',
                               dest='outdir',
                               help='Output directory',
                               type=str,
                               default=OUTDIR,
    )
    parent_parser.add_argument('--verbose', '-v', 
                               dest='verbose',
                               help='Turn on verbose output',
                               default=False, 
                               action='store_true',
    )

    # Main options 
    main_parser = argparse.ArgumentParser(add_help=False)

    main_parser.add_argument('--kc', '--kepcorr', '-k', 
                             dest='kepcorr',
                             help='Turn on the Kepler short-cadence artefact correction routine',
                             default=False, 
                             action='store_true',
    )
    main_parser.add_argument('--ofa', '--of_actual',
                             metavar='n',
                             dest='of_actual',
                             help='The oversampling factor (OF) of the input PS',
                             type=int,
                             default=None,
    )
    main_parser.add_argument('--ofn', '--of_new',
                             metavar='n',
                             dest='of_new',
                             help='The OF to be used for the first iteration',
                             type=int,
                             default=None,
    )
    main_parser.add_argument('--save',
                             dest='save',
                             help='Do not save output files/figures',
                             default=True, 
                             action='store_false',
    )
    main_parser.add_argument('--show', '-s', 
                             dest='show',
                             help='Display figures',
                             default=False, 
                             action='store_true',
    )
    main_parser.add_argument('--star', '--stars',
                             metavar='star',
                             dest='stars',
                             help='List of stars to process',
                             type=str,
                             nargs='*',
                             default=None,
    )

    # CLI relevant for finding power excess
    excess = main_parser.add_argument_group('(CRUDE) EXCESS FIT')

    excess.add_argument('--bin', '--binning',
                        metavar='value',  
                        dest='binning', 
                        help='Binning interval for PS (in muHz)',
                        default=0.005, 
                        type=float,
    )
    excess.add_argument('--bm', '--mode', '--bmode',
                        metavar='mode',
                        choices=["mean", "median", "gaussian"],
                        dest='mode',
                        help='Binning mode',
                        default='mean',
                        type=str,
    )
    excess.add_argument('--sw', '--smoothwidth',
                        metavar='value', 
                        dest='smooth_width',
                        help='Box filter width (in muHz) for smoothing the PS',
                        default=50.0,
                        type=float,
    )
    excess.add_argument('--lx', '--lowerx', 
                        metavar='freq', 
                        dest='lower_ex',
                        help='Lower frequency limit of PS',
                        nargs='*',
                        default=None,
                        type=float,
    )
    excess.add_argument('--ux', '--upperx', 
                        metavar='freq', 
                        dest='upper_ex',
                        help='Upper frequency limit of PS',
                        nargs='*',
                        default=None,
                        type=float,
    )
    excess.add_argument('--step', '--steps', 
                        metavar='value', 
                        dest='step', 
                        default=0.25,
                        type=float, 
    )
    excess.add_argument('--trials', '--ntrials',
                        metavar='n', 
                        dest='n_trials',
                        default=3, 
                        type=int,
    )

    # CLI relevant for background fitting
    background = main_parser.add_argument_group('BACKGROUND-RELATED')

    background.add_argument('--lb', '--lowerb', 
                            metavar='freq', 
                            dest='lower_bg',
                            help='Lower frequency limit of PS',
                            nargs='*',
                            default=None,
                            type=float,
    )
    background.add_argument('--ub', '--upperb', 
                            metavar='freq', 
                            dest='upper_bg',
                            help='Upper frequency limit of PS',
                            nargs='*',
                            default=None,
                            type=float,
    )
    background.add_argument('--iw', '--indwidth',
                            metavar='value', 
                            dest='ind_width', 
                            help='Width of binning for PS [in muHz]',
                            default=20., 
                            type=float,
    )
    background.add_argument('--bf', '--box', '--boxfilter',
                            metavar='value', 
                            dest='box_filter',
                            help='Box filter width [in muHz] for plotting the PS',
                            default=1.0,
                            type=float,
    )
    background.add_argument('--rms', '--nrms', 
                            metavar='n', 
                            dest='n_rms', 
                            help='Number of points to estimate the amplitude of red-noise component(s)',
                            default=20, 
                            type=int,
    )
    background.add_argument('--laws', '--nlaws', 
                            metavar='n', 
                            dest='n_laws', 
                            help='Force number of red-noise component(s)',
                            default=None, 
                            type=int,
    )
    background.add_argument('--ab', '-ab', 
                            dest='ab',
                            help='Use the {a,b} parametrization for Harvey models',
                            default=False, 
                            action='store_true',
    )
    background.add_argument('--use', 
                            metavar='metric', 
                            dest='use', 
                            help="Which model metric to use, choices=['bic','aic']",
                            default='bic', 
                            type=str,
    )

    # CLI relevant for the global parameters
    numax = main_parser.add_argument_group('NUMAX-RELATED')
    numax.add_argument('--sm', '--smpar',
                       metavar='value', 
                       dest='sm_par',
                       help='Value of smoothing parameter to estimate smoothed numax (typically between 1-4).',
                       default=None, 
                       type=float,
    )
    numax.add_argument('--numax',
                       metavar='value',
                       dest='numax',
                       help='Skip find excess module and force numax',
                       nargs='*',
                       default=None,
                       type=float,
    )
    numax.add_argument('--lp', '--lowerp', 
                       metavar='freq', 
                       dest='lower_ps',
                       help='Lower frequency limit for zoomed in PS',
                       nargs='*',
                       default=None,
                       type=float,
    )
    numax.add_argument('--up', '--upperp', 
                       metavar='freq', 
                       dest='upper_ps',
                       help='Upper frequency limit for zoomed in PS',
                       nargs='*',
                       default=None,
                       type=float,
    )
    numax.add_argument('--ew', '--exwidth',
                       metavar='value', 
                       dest='width',
                       help='Fractional value of width to use for power excess, where width is computed using a solar scaling relation.',
                       default=1.0,
                       type=float,
    )

    dnu = main_parser.add_argument_group('DNU-RELATED')
    dnu.add_argument('--dnu',
                     metavar='value',
                     dest='dnu',
                     help='Brute force method to provide value for dnu',
                     nargs='*',
                     type=float,
                     default=None, 
    )
    dnu.add_argument('--sp', '--smoothps',
                     metavar='value', 
                     dest='smooth_ps',
                     help='Box filter width [in muHz] of PS for ACF', 
                     default=2.5,
                     type=float,
    )
    dnu.add_argument('--peak', '--peaks', '--npeaks',
                     metavar='n', 
                     dest='n_peaks', 
                     help='Number of peaks to fit in the ACF',
                     default=5, 
                     type=int,
    )
    dnu.add_argument('--thresh', '--threshold',
                     metavar='value', 
                     dest='threshold',
                     help='Fractional value of FWHM to use for ACF',
                     default=1.0,
                     type=float,
    )

    echelle = main_parser.add_argument_group('ECHELLE-RELATED')
    echelle.add_argument('--cv', '--value',
                         metavar='value', 
                         dest='clip_value',
                         help='Clip value to use for echelle diagram (ED)',
                         default=None, 
                         type=float,
    )
    echelle.add_argument('--ce', '--clipech', 
                         dest='clip_ech',
                         help='Disable the auto-clipping of high peaks in the ED',
                         default=True, 
                         action='store_false',
    )
    echelle.add_argument('--le', '--lowere', 
                         metavar='freq', 
                         dest='lower_ech',
                         help='Lower frequency limit of folded PS to whiten mixed modes',
                         nargs='*',
                         default=None,
                         type=float,
    )
    echelle.add_argument('--ue', '--uppere', 
                         metavar='freq', 
                         dest='upper_ech',
                         help='Upper frequency limit of folded PS to whiten mixed modes',
                         nargs='*',
                         default=None,
                         type=float,
    )
    echelle.add_argument('--hey',
                         dest='hey', 
                         help="Use Daniel Hey's plugin for echelle",
                         default=False, 
                         action='store_true',
    )
    echelle.add_argument('--ie', '-interpech', '--interpech',
                         dest='interp_ech',
                         help='Turn on the interpolation of the output ED',
                         default=False,
                         action='store_true',
    )
    echelle.add_argument('--se', '--smoothech',
                         metavar='value', 
                         dest='smooth_ech',
                         help='Smooth ED using a box filter [in muHz]',
                         default=None,
                         type=float,
    )
    echelle.add_argument('--xe', '--xech', '--nacross',
                         metavar='n', 
                         dest='n_across',
                         help='Resolution for the x-axis of the ED',
                         default=50,
                         type=int, 
    )
    echelle.add_argument('--ye', '--yech', '--ndown',
                         metavar='n', 
                         dest='n_down',
                         help='The number of orders to plot on the ED y-axis',
                         default=5,
                         type=int,
    )

    mcmc = main_parser.add_argument_group('MCMC PARAMETERS')
    mcmc.add_argument('--mc', '--iter', '--mciter', 
                      metavar='n', 
                      dest='mc_iter', 
                      help='Number of Monte-Carlo iterations',
                      default=1, 
                      type=int,
    )
    mcmc.add_argument('--samples', 
                      dest='samples',
                      help='Save samples from the Monte-Carlo sampling',
                      default=False, 
                      action='store_true',
    )

    sub_parser = parser.add_subparsers(title='pySYD modes', dest='modes')

    # Setting up
    parser_setup = sub_parser.add_parser('setup', help='Easy setup of relevant directories and files',
                                         parents=[parent_parser], formatter_class=argparse.MetavarTypeHelpFormatter)
    parser_setup.set_defaults(func=pipeline.setup)

    # Load data in for a target
    parser_load = sub_parser.add_parser('load', help='Load in data for a given target',
                                       parents=[parent_parser, main_parser], formatter_class=argparse.MetavarTypeHelpFormatter)

    parser_load.set_defaults(func=pipeline.load)

    # Running pySYD in regular mode
    parser_run = sub_parser.add_parser('run', help='Run the main pySYD pipeline',
                                       parents=[parent_parser, main_parser], formatter_class=argparse.MetavarTypeHelpFormatter)

    parser_run.set_defaults(func=pipeline.run)

    # Run pySYD in parallel
    parser_parallel = sub_parser.add_parser('parallel', help='Run pySYD in parallel',
                                            parents=[parent_parser, main_parser], formatter_class=argparse.MetavarTypeHelpFormatter)
    parser_parallel.add_argument('-nt', '--nt', '-nthread', '--nthread', '-nthreads', '--nthreads',
                                 metavar='n', 
                                 dest='n_threads',
                                 help='Number of processes to run in parallel',
                                 type=int,
                                 default=0,
    )

    parser_parallel.set_defaults(func=pipeline.parallel)

    args = parser.parse_args()
    args.func(args)



if __name__ == '__main__':

    main()


import os

from .functions import *
from .pipeline import *
from .models import *
from .target import *
from .plots import *
from .utils import *

__all__ = ['cli', 'functions', 'pipeline', 'models', 'target', 'plots', 'utils']

__version__ = '2.1.1'

_ROOT = os.path.abspath(os.getcwd())
TODODIR = os.path.join(_ROOT, 'info', 'todo.txt')
INFODIR = os.path.join(_ROOT, 'info', 'star_info.csv')
INPDIR = os.path.join(_ROOT, 'data')
OUTDIR = os.path.join(_ROOT, 'results')


import re
import setuptools

def get_property(prop, project):
    result = re.search(r'{}\s*=\s*[\'"]([^\'"]*)[\'"]'.format(prop),
                       open(project+'/__init__.py').read())
    return result.group(1)

with open("README.md", "r", encoding="utf-8") as file:
    long_description = file.read()

reqs = []
for line in open('requirements.txt', 'r').readlines():
    reqs.append(line)

setuptools.setup(
    name="pysyd",
    version=get_property('__version__', 'pysyd'),
    license="MIT",
    author="Ashley Chontos",
    author_email="achontos@hawaii.edu",
    description="Automated measurements of global asteroseismic parameters",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ashleychontos/pysyd",
    project_urls={
        "Documentation": "https://pysyd.readthedocs.io",
        "Source": "https://github.com/ashleychontos/pySYD",
        "Bug Tracker": "https://github.com/ashleychontos/pySYD/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=reqs,
    packages=setuptools.find_packages(),
    entry_points={'console_scripts':['pysyd=pysyd.cli:main']},
    python_requires=">=3.6",
)
